from pydantic import BaseModel
from typing import Union
import copy
import functools
import inspect


class AgentContext(BaseModel):
    """
    Agent Context
    """

    ...


class BlockContext(BaseModel):
    """
    Block Context
    """

    ...


class DotDict(dict):
    """
    A dictionary that supports dot notation access to its items.

    - **Description**:
        - Extends the standard dictionary to allow attribute-style access
        - Example: d = DotDict({'foo': 'bar'}); d.foo == 'bar'
        - Supports merging with other DotDict instances
        - Maintains reference to original dictionaries when merged

    - **Args**:
        - Same as dict constructor

    - **Returns**:
        - A dictionary with attribute-style access
    """

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        if key in self:
            del self[key]

    def __init__(self, *args, **kwargs):
        super(DotDict, self).__init__(*args, **kwargs)
        # Convert nested dictionaries to DotDict
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = DotDict(value)
            elif isinstance(value, list):
                self[key] = [
                    DotDict(item) if isinstance(item, dict) else item for item in value
                ]

    def merge(self, other):
        """
        Merges another DotDict into this one.

        - **Description**:
            - Merges another DotDict into this one
            - Maintains references to original dictionaries
            - Updates are synchronized with original dictionaries

        - **Args**:
            - `other` (DotDict): The DotDict to merge with

        - **Returns**:
            - `self` (DotDict): The merged DotDict
        """
        if not isinstance(other, DotDict):
            other = DotDict(other)

        for key, value in other.items():
            if (
                key in self
                and isinstance(self[key], DotDict)
                and isinstance(value, DotDict)
            ):
                self[key].merge(value)
            else:
                self[key] = value
        return self

    def __or__(self, other):
        """
        Implements the | operator for merging DotDicts.

        - **Description**:
            - Allows using the | operator to merge DotDicts
            - Example: c = a | b

        - **Args**:
            - `other` (DotDict): The DotDict to merge with

        - **Returns**:
            - `DotDict`: A new merged DotDict
        """
        result = DotDict(self)
        return result.merge(other)

    def __ior__(self, other):
        """
        Implements the |= operator for in-place merging.

        - **Description**:
            - Allows using the |= operator for in-place merging
            - Example: a |= b

        - **Args**:
            - `other` (DotDict): The DotDict to merge with

        - **Returns**:
            - `self` (DotDict): The updated DotDict
        """
        return self.merge(other)


def context_to_dot_dict(context: Union[AgentContext, BlockContext]) -> DotDict:
    """
    Converts an AgentContext to a DotDict for dot notation access.

    - **Description**:
        - Converts a Pydantic AgentContext model to a dictionary with dot notation access
        - Preserves nested structure while enabling attribute-style access

    - **Args**:
        - `context` (AgentContext): The AgentContext to convert

    - **Returns**:
        - `dot_dict` (DotDict): A dictionary that supports attribute-style access
    """
    # Convert to dict first using model_dump() method
    context_dict = context.model_dump()
    # Convert the dict to a DotDict
    return DotDict(context_dict)


# Decorator to automatically deepcopy DotDict arguments
def auto_deepcopy_dotdict(func):
    """
    Decorator that automatically deep copies DotDict arguments.

    - **Description**:
        - Wraps a function to automatically deep copy any DotDict arguments
        - Prevents modifications to DotDict arguments from affecting the original

    - **Args**:
        - `func` (callable): The function to decorate

    - **Returns**:
        - `wrapper` (callable): The decorated function
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Deep copy DotDict positional arguments
        new_args = []
        for arg in args:
            if isinstance(arg, DotDict):
                new_args.append(copy.deepcopy(arg))
            else:
                new_args.append(arg)

        # Deep copy DotDict keyword arguments
        new_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, DotDict):
                new_kwargs[key] = copy.deepcopy(value)
            else:
                new_kwargs[key] = value

        return func(*new_args, **new_kwargs)

    return wrapper


# Apply the decorator automatically to all functions in a module
def apply_auto_deepcopy_to_module(module):
    """
    Applies the auto_deepcopy_dotdict decorator to all functions in a module.

    - **Description**:
        - Scans a module for functions and applies the decorator to each
        - Makes all functions in the module automatically deep copy DotDict arguments

    - **Args**:
        - `module` (module): The module to process

    - **Returns**:
        - None
    """
    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) and not name.startswith("_"):
            setattr(module, name, auto_deepcopy_dotdict(obj))



from ..llm import LLM, LLMConfig
from .messager import Message, MessageKind


__all__ = [
    "MessageInterceptor",
]


class MessageInterceptor:
    """
    A class to intercept and process messages based on configured rules.
    """

    def __init__(
        self,
        llm_config: list[LLMConfig],
    ) -> None:
        """
        Initialize the MessageInterceptor with optional configuration.

        - **Args**:
            - `llm_config` (LLMConfig): Configuration dictionary for initializing the LLM instance. Defaults to None.
        """
        self._llm = LLM(llm_config)
        # round related
        self.validation_dict: dict[Message, bool] = {}
        self._supervisor = None

    async def set_supervisor(self, supervisor):
        self._supervisor = supervisor

    # Property accessors
    @property
    def supervisor(self):
        if self._supervisor is None:
            raise RuntimeError("Supervisor not set")
        return self._supervisor

    @property
    def llm(self) -> LLM:
        """
        Access the Large Language Model instance.

        - **Description**:
            - Provides access to the internal LLM instance. Raises an error if accessed before assignment.

        - **Raises**:
            - `RuntimeError`: If accessed before setting the LLM.

        - **Returns**:
            - `LLM`: The Large Language Model instance.
        """
        if self._llm is None:
            raise RuntimeError("LLM access before assignment, please `set_llm` first!")
        return self._llm

    # Message forwarding related methods
    async def forward(self, messages: list[Message]) -> list[Message]:
        # reset round related variables
        new_round_validation_dict, persuasion_messages = await self.supervisor.forward(
            messages
        )
        result_messages: list[Message] = []
        for msg in messages:
            if msg.kind == MessageKind.AGENT_CHAT:
                is_valid = new_round_validation_dict.get(msg, True)
                if is_valid:
                    result_messages.append(msg)
                else:
                    # add message to from_id
                    result_messages.append(
                        Message(
                            from_id=msg.to_id,
                            to_id=msg.from_id,
                            kind=MessageKind.AGENT_CHAT,
                            payload={"content": "Message sent failed"},
                            day=msg.day,
                            t=msg.t,
                        )
                    )
            else:
                result_messages.append(msg)
        result_messages.extend(persuasion_messages)
        return result_messages



"""Environment"""

from .sim import PersonService
from .environment import (
    Environment,
    EnvironmentStarter,
    EnvironmentConfig,
)
from .mapdata import MapData, MapConfig
from .economy import EconomyClient

__all__ = [
    "Environment",
    "EnvironmentStarter",
    "EnvironmentConfig",
    "MapData",
    "MapConfig",
    "PersonService",
    "EconomyClient",
]



from fastapi import APIRouter

from .experiment import router as experiment_router
from .agent import router as agent_dialog_router
from .survey import router as survey_router
from .experiment_runner import router as experiment_runner_router
from .config import router as config_router
from .agent_profiles import router as agent_profiles_router
from .agent_template import router as agent_template_router

__all__ = ["api_router"]

# Main API router
api_router = APIRouter(prefix="/api")

# Include sub-routers
api_router.include_router(experiment_router)
api_router.include_router(agent_dialog_router)
api_router.include_router(survey_router)
api_router.include_router(experiment_runner_router)
api_router.include_router(config_router)
api_router.include_router(agent_profiles_router)
api_router.include_router(agent_template_router)


"""
Agent Message System including message interceptor and messager.

- **Message Interceptor**: Intercepts messages from the message queue and processes them.
- **Messager**: Sends and receives messages.
"""

from .message_interceptor import MessageInterceptor
from .messager import Messager, Message, MessageKind

__all__ = [
    "Message",
    "Messager",
    "MessageKind",
    "MessageInterceptor",
]



"""
AgentSociety Commercial Features

This package contains commercial-only features including:
- Casdoor authentication
- Kubernetes executor
- Billing system
- Payment integration

When this package is removed or not available, the system will
automatically fall back to open-source alternatives.
"""

import logging
from typing import Any, Dict

logger = logging.getLogger(__name__)

def is_available() -> bool:
    """Check if commercial features are available"""
    try:
        # Try to import a core commercial module to verify availability
        from . import auth
        _ = auth
        return True
    except ImportError:
        return False

def get_auth_provider(config: Dict[str, Any]):
    """Get commercial authentication provider"""
    if not is_available():
        return None
    
    try:
        from .auth.casdoor import get_casdoor_auth
        return get_casdoor_auth(config)
    except ImportError as e:
        logger.warning(f"Failed to load commercial auth provider: {e}")
        return None

def get_kubernetes_executor(config: Dict[str, Any]):
    """Get commercial Kubernetes executor"""
    if not is_available():
        return None
    
    try:
        from .executor import get_kubernetes_executor
        return get_kubernetes_executor(config)
    except ImportError as e:
        logger.warning(f"Failed to load commercial Kubernetes executor: {e}")
        return None

def get_billing_system(config: Dict[str, Any]):
    """Get commercial billing system"""
    if not is_available():
        return None
    
    try:
        from .billing.system import get_billing_system
        return get_billing_system(config)
    except ImportError as e:
        logger.warning(f"Failed to load commercial billing system: {e}")
        return None 


from typing import Any
import json_repair
from openai.types.chat import ChatCompletionToolParam

from ..logger import get_logger
from ..memory import Memory
from .block import Block
from .context import DotDict
from .prompt import FormatPrompt
from .toolbox import AgentToolbox

DISPATCHER_PROMPT = """
Based on the task information (which describes the needs of the user), select the most appropriate block to handle the task.
Each block has its specific functionality as described in the function schema.
        
Task information:
${context.current_intention}
"""


class BlockDispatcher:
    """Orchestrates task routing between registered processing blocks.

    Attributes:
        toolbox: AgentToolbox
        blocks: Registry of available processing blocks (name -> Block mapping)
        prompt: Formatted prompt template for LLM instructions
    """

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        selection_prompt: str = DISPATCHER_PROMPT,
    ):
        """Initialize dispatcher with LLM interface.

        Args:
            llm: Language model for block selection decisions
        """
        self.toolbox = toolbox
        self.memory = agent_memory
        self.blocks: dict[str, Block] = {}
        self.dispatcher_prompt = FormatPrompt(selection_prompt, memory=self.memory)

    def register_dispatcher_prompt(self, dispatcher_prompt: str) -> None:
        """Register a dispatcher prompt.

        Args:
            dispatcher_prompt: Dispatcher prompt
        """
        self.dispatcher_prompt = FormatPrompt(dispatcher_prompt, memory=self.memory)

    def register_blocks(self, blocks: list[Block]) -> None:
        """Register multiple processing blocks for dispatching.

        Args:
            blocks: List of Block instances to register
        """
        for block in blocks:
            block_name = block.__class__.__name__.lower()
            self.blocks[block_name] = block

    def _get_function_schema(self) -> ChatCompletionToolParam:
        """
        Generate LLM function calling schema describing available blocks.

        - **Description**:
            - Creates a schema for the LLM to select appropriate blocks or indicate no suitable block exists.

        - **Returns**:
            - `ChatCompletionToolParam`: Function schema dictionary compatible with OpenAI-style function calling
        """
        # create block descriptions
        block_descriptions = {
            name: block.description for name, block in self.blocks.items()
        }

        return {
            "type": "function",
            "function": {
                "name": "select_block",
                "description": "Select the most appropriate block based on the step information, or indicate no suitable block exists",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "block_name": {
                            "type": "string",
                            "enum": list(self.blocks.keys()) + ["no_suitable_block"],
                            "description": f"Available blocks and their descriptions: {block_descriptions}. Use 'no_suitable_block' if none of the blocks are appropriate for the given intention.",
                        },
                        "reason": {
                            "type": "string",
                            "description": "Explanation for why the selected block is appropriate or why no suitable block exists",
                        },
                    },
                    "required": ["block_name", "reason"],
                },
            },
        }

    async def dispatch(self, context: DotDict) -> Block | None:
        """
        Route a task step to the most appropriate processing block.

        - **Description**:
            - Uses LLM to select the best block for handling the given task intention.
            - Can return None if no suitable block is found.

        - **Args**:
            - `intention` (str): Intention of the task

        - **Returns**:
            - `Block | None`: Selected Block instance for handling the task, or None if no suitable block exists
        """
        try:
            function_schema = self._get_function_schema()
            await self.dispatcher_prompt.format(context=context)

            # Call LLM with tools schema
            response = await self.toolbox.llm.atext_request(
                self.dispatcher_prompt.to_dialog(),
                tools=[function_schema],
                tool_choice={"type": "function", "function": {"name": "select_block"}},
            )
            function_args: Any = json_repair.loads(
                response.choices[0].message.tool_calls[0].function.arguments
            )
            selected_block = function_args.get("block_name")
            reason = function_args.get("reason", "No reason provided")

            if selected_block == "no_suitable_block":
                get_logger().debug(
                    f"No suitable block found for intention. Reason: {reason}"
                )
                return None

            if selected_block not in self.blocks:
                raise ValueError(
                    f"Selected block '{selected_block}' not found in registered blocks"
                )

            get_logger().debug(
                f"Dispatched intention to block: {selected_block}. Reason: {reason}"
            )
            return self.blocks[selected_block]

        except Exception as e:
            get_logger().warning(f"Failed to dispatch block: {e}")
            return None



import math
import random
from typing import Optional

import json_repair
import numpy as np
from pydantic import Field

from ...agent import (
    Block,
    FormatPrompt,
    BlockParams,
    DotDict,
    BlockContext,
    AgentToolbox,
)
from ...logger import get_logger
from ...memory import Memory
from ...agent.dispatcher import BlockDispatcher
from ..sharing_params import SocietyAgentBlockOutput
from .utils import clean_json_response

# Prompt templates for LLM interactions
PLACE_TYPE_SELECTION_PROMPT = """
As an intelligent decision system, please determine the type of place the user needs to visit based on their input requirement.
User Plan: {plan}
User requirement: {intention}
Other information: 
-------------------------
{other_info}
-------------------------
Your output must be a single selection from {poi_category} without any additional text or explanation.

Please response in json format (Do not return any other text), example:
{{
    "place_type": "shopping"
}}
"""

PLACE_SECOND_TYPE_SELECTION_PROMPT = """
As an intelligent decision system, please determine the type of place the user needs to visit based on their input requirement.
User Plan: {plan}
User requirement: {intention}
Other information: 
-------------------------
{other_info}
-------------------------

Your output must be a single selection from {poi_category} without any additional text or explanation.

Please response in json format (Do not return any other text), example:
{{
    "place_type": "shopping"
}}
"""

PLACE_ANALYSIS_PROMPT = """
As an intelligent analysis system, please determine the type of place the user needs to visit based on their input requirement.
User Plan: {plan}
User requirement: {intention}
Other information: 
-------------------------
{other_info}
-------------------------

Your output must be a single selection from {place_list} without any additional text or explanation.

Please response in json format (Do not return any other text), example:
{{
    "place_type": "home"
}}
"""

RADIUS_PROMPT = """As an intelligent decision system, please determine the maximum travel radius (in meters) based on the current emotional state.

Current weather: ${context.weather}
Current temperature: ${context.temperature}
Your current emotion: ${context.current_emotion}
Your current thought: ${context.current_thought}
Other information: 
-------------------------
${context.other_information}
-------------------------

Please analyze how these emotions would affect travel willingness and return only a single integer number between 3000-200000 representing the maximum travel radius in meters. A more positive emotional state generally leads to greater willingness to travel further.

Please response in json format (Do not return any other text), example:
{{
    "radius": 10000
}}
"""


def gravity_model(pois):
    """
    Calculate selection probabilities for POIs using a gravity model.

    The model considers both distance decay (prefer closer locations)
    and spatial density (avoid overcrowded areas). Distances are grouped
    into 1km bins up to 10km, with POIs beyond 10km in a 'more' category.

    Args:
        pois: List of POI tuples containing (poi_data, distance)

    Returns:
        List of tuples: (name, id, normalized_weight, distance)
        with selection probabilities based on gravity model
    """
    # Initialize distance bins
    pois_Dis = {f"{d}k": [] for d in range(1, 11)}
    pois_Dis["more"] = []

    # Categorize POIs into distance bins
    for poi in pois:
        classified = False
        for d in range(1, 11):
            if (d - 1) * 1000 <= poi[1] < d * 1000:
                pois_Dis[f"{d}k"].append(poi)
                classified = True
                break
        if not classified:
            pois_Dis["more"].append(poi)

    res = []
    distanceProb = []
    # Calculate weights for each POI
    for poi in pois:
        for d in range(1, 11):
            if (d - 1) * 1000 <= poi[1] < d * 1000:
                n = len(pois_Dis[f"{d}k"])
                # Calculate ring area between (d-1)km and d km
                S = math.pi * ((d * 1000) ** 2 - ((d - 1) * 1000) ** 2)
                density = n / S  # POIs per square meter
                distance = max(poi[1], 1)  # Avoid division by zero

                # Inverse square distance decay combined with density
                weight = density / (distance**2)
                res.append((poi[0]["name"], poi[0]["id"], weight, distance))
                distanceProb.append(1 / math.sqrt(distance))
                break

    # Normalize probabilities and sample
    distanceProb = np.array(distanceProb)
    distanceProb /= distanceProb.sum()

    # Randomly sample 50 candidates weighted by distance probabilities
    sample_indices = np.random.choice(len(res), size=50, p=distanceProb)
    sampled_pois = [res[i] for i in sample_indices]

    # Normalize weights for final selection
    total_weight = sum(item[2] for item in sampled_pois)
    return [
        (item[0], item[1], item[2] / total_weight, item[3]) for item in sampled_pois
    ]


class PlaceSelectionBlock(Block):
    """
    Block for selecting destinations based on user intention.

    Implements a two-stage selection process:
    1. Select primary POI category (e.g., 'shopping')
    2. Select sub-category (e.g., 'bookstore')
    Uses LLM for decision making with fallback to random selection.

    Configurable Fields:
        search_limit: Max number of POIs to retrieve from map service
    """

    name = "PlaceSelectionBlock"
    description = "Selects destinations for unknown locations (excluding home/work)"

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        search_limit: int = 50,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.typeSelectionPrompt = FormatPrompt(PLACE_TYPE_SELECTION_PROMPT)
        self.secondTypeSelectionPrompt = FormatPrompt(
            PLACE_SECOND_TYPE_SELECTION_PROMPT
        )
        self.radiusPrompt = FormatPrompt(
            RADIUS_PROMPT,
            memory=agent_memory,
        )
        self.search_limit = search_limit  # Default config value

    async def forward(self, context: DotDict):
        """Execute the destination selection workflow"""
        # Stage 1: Select primary POI category
        poi_cate = self.environment.get_poi_cate()
        await self.typeSelectionPrompt.format(
            plan=context["plan_context"]["plan"],
            intention=context["current_step"]["intention"],
            poi_category=list(poi_cate.keys()),
            other_info=self.environment.environment.get("other_information", "None"),
        )
        try:
            # LLM-based category selection
            levelOneType = await self.llm.atext_request(
                self.typeSelectionPrompt.to_dialog(),
                response_format={"type": "json_object"},
            )
            levelOneType = json_repair.loads(clean_json_response(levelOneType))["place_type"] # type: ignore
            sub_category = poi_cate[levelOneType]
        except Exception as e:
            get_logger().warning(f"MobilityBlock: Level 1 selection failed: {e}")
            levelOneType = random.choice(list(poi_cate.keys()))
            sub_category = poi_cate[levelOneType]

        # Stage 2: Select sub-category
        try:
            await self.secondTypeSelectionPrompt.format(
                plan=context["plan_context"]["plan"],
                intention=context["current_step"]["intention"],
                poi_category=sub_category,
                other_info=self.environment.environment.get(
                    "other_information", "None"
                ),
            )
            levelTwoType = await self.llm.atext_request(
                self.secondTypeSelectionPrompt.to_dialog(),
                response_format={"type": "json_object"},
            )
            levelTwoType = json_repair.loads(clean_json_response(levelTwoType))["place_type"] # type: ignore
        except Exception as e:
            get_logger().warning(f"MobilityBlock: Level 2 selection failed: {e}")
            levelTwoType = random.choice(sub_category)

        # Get travel radius from LLM
        try:
            await self.radiusPrompt.format(context=context)
            radius = await self.llm.atext_request(
                self.radiusPrompt.to_dialog(), response_format={"type": "json_object"}
            )
            radius = int(json_repair.loads(radius)["radius"]) # type: ignore
        except Exception as e:
            get_logger().warning(f"MobilityBlock: Radius selection failed: {e}")
            radius = 10000  # Default 10km

        # Query and select POI
        xy = (await self.memory.status.get("position"))["xy_position"]
        center = (xy["x"], xy["y"])
        pois = self.environment.map.query_pois(
            center=center,
            category_prefix=levelTwoType,
            radius=radius,
            limit=self.search_limit,
        )

        if pois:
            pois = gravity_model(pois)
            probabilities = [item[2] for item in pois]
            selected = np.random.choice(len(pois), p=probabilities)
            next_place = (pois[selected][0], pois[selected][1])
        else:  # Fallback random selection
            all_pois = self.environment.map.get_all_pois()
            next_place = random.choice(all_pois)
            next_place = (next_place["name"], next_place["id"])

        context["next_place"] = next_place
        node_id = await self.memory.stream.add(
            topic="mobility",
            description=f"For {context['current_step']['intention']}, selected: {next_place}",
        )
        return {
            "success": True,
            "evaluation": f"Selected destination: {next_place}",
            "consumed_time": 5,
            "node_id": node_id,
        }


class MoveBlock(Block):
    """Block for executing mobility operations (home/work/other)"""

    name = "MoveBlock"
    description = "Executes mobility operations between locations"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.placeAnalysisPrompt = FormatPrompt(PLACE_ANALYSIS_PROMPT)

    async def forward(self, context: DotDict):
        agent_id = await self.memory.status.get("id")
        place_knowledge = await self.memory.status.get("location_knowledge")
        known_places = list(place_knowledge.keys())
        places = ["home", "workplace"] + known_places + ["other"]
        await self.placeAnalysisPrompt.format(
            plan=context["plan_context"]["plan"],
            intention=context["current_step"]["intention"],
            place_list=places,
            other_info=self.environment.environment.get("other_information", "None"),
        )
        response = await self.llm.atext_request(
            self.placeAnalysisPrompt.to_dialog(),
            response_format={"type": "json_object"},
        )  #
        try:
            response = clean_json_response(response)
            response = json_repair.loads(response)["place_type"] # type: ignore
        except Exception:
            get_logger().warning(
                f"MobilityBlock: Place Analysis: wrong type of place, raw response: {response}"
            )
            response = "home"
        if response == "home":
            # go back home
            home = await self.memory.status.get("home")
            home = home["aoi_position"]["aoi_id"]
            nowPlace = await self.memory.status.get("position")
            node_id = await self.memory.stream.add(
                topic="mobility",
                description="I returned home",
            )
            if (
                "aoi_position" in nowPlace
                and nowPlace["aoi_position"]["aoi_id"] == home
            ):
                return {
                    "success": True,
                    "evaluation": "Successfully returned home (already at home)",
                    "to_place": home,
                    "consumed_time": 0,
                    "node_id": node_id,
                }
            await self.environment.set_aoi_schedules(
                person_id=agent_id,
                target_positions=home,
            )
            number_poi_visited = await self.memory.status.get("number_poi_visited")
            number_poi_visited += 1
            await self.memory.status.update("number_poi_visited", number_poi_visited)
            return {
                "success": True,
                "evaluation": "Successfully returned home",
                "to_place": home,
                "consumed_time": 45,
                "node_id": node_id,
            }
        elif response == "workplace":
            # back to workplace
            work = await self.memory.status.get("work")
            work = work["aoi_position"]["aoi_id"]
            nowPlace = await self.memory.status.get("position")
            node_id = await self.memory.stream.add(
                topic="mobility",
                description="I went to my workplace",
            )
            if (
                "aoi_position" in nowPlace
                and nowPlace["aoi_position"]["aoi_id"] == work
            ):
                return {
                    "success": True,
                    "evaluation": "Successfully reached the workplace (already at the workplace)",
                    "to_place": work,
                    "consumed_time": 0,
                    "node_id": node_id,
                }
            await self.environment.set_aoi_schedules(
                person_id=agent_id,
                target_positions=work,
            )
            number_poi_visited = await self.memory.status.get("number_poi_visited")
            number_poi_visited += 1
            await self.memory.status.update("number_poi_visited", number_poi_visited)
            return {
                "success": True,
                "evaluation": "Successfully reached the workplace",
                "to_place": work,
                "consumed_time": 45,
                "node_id": node_id,
            }
        elif response in known_places:
            the_place = place_knowledge[response]["id"]
            nowPlace = await self.memory.status.get("position")
            node_id = await self.memory.stream.add(
                topic="mobility",
                description=f"I went to {response}",
            )
            if (
                "aoi_position" in nowPlace
                and nowPlace["aoi_position"]["aoi_id"] == the_place
            ):
                return {
                    "success": True,
                    "evaluation": f"Successfully reached {response} (already at {response})",
                    "to_place": the_place,
                    "consumed_time": 0,
                    "node_id": node_id,
                }
            await self.environment.set_aoi_schedules(
                person_id=agent_id,
                target_positions=the_place,
            )
            number_poi_visited = await self.memory.status.get("number_poi_visited")
            number_poi_visited += 1
            await self.memory.status.update("number_poi_visited", number_poi_visited)
            return {
                "success": True,
                "evaluation": f"Successfully reached {response}",
                "to_place": the_place,
                "consumed_time": 45,
                "node_id": node_id,
            }
        else:
            # move to other places
            next_place = context.get("next_place", None)
            nowPlace = await self.memory.status.get("position")
            node_id = await self.memory.stream.add(
                topic="mobility",
                description=f"I went to {next_place}",
            )
            if next_place is not None:
                await self.environment.set_aoi_schedules(
                    person_id=agent_id,
                    target_positions=next_place[1],
                )
            else:
                aois = self.environment.map.get_all_aois()
                while True:
                    r_aoi = random.choice(aois)
                    if len(r_aoi["poi_ids"]) > 0:
                        r_poi = random.choice(r_aoi["poi_ids"])
                        break
                poi = self.environment.map.get_poi(r_poi)
                next_place = (poi["name"], poi["aoi_id"])
                await self.environment.set_aoi_schedules(
                    person_id=agent_id,
                    target_positions=next_place[1],
                )
            number_poi_visited = await self.memory.status.get("number_poi_visited")
            number_poi_visited += 1
            await self.memory.status.update("number_poi_visited", number_poi_visited)
            return {
                "success": True,
                "evaluation": f"Successfully reached the destination: {next_place}",
                "to_place": next_place[1],
                "consumed_time": 45,
                "node_id": node_id,
            }


class MobilityNoneBlock(Block):
    """
    MobilityNoneBlock
    """

    name = "MobilityNoneBlock"
    description = "Handles other mobility operations"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )

    async def forward(self, context: DotDict):
        """Log completion without action"""
        node_id = await self.memory.stream.add(
            topic="mobility",
            description=f"I finished {context['current_step']['intention']}",
        )
        return {
            "success": True,
            "evaluation": f"Finished executing {context['current_step']['intention']}",
            "consumed_time": 0,
            "node_id": node_id,
        }


class MobilityBlockParams(BlockParams):
    # PlaceSelection
    radius_prompt: str = Field(
        default=RADIUS_PROMPT, description="Used to determine the maximum travel radius"
    )
    search_limit: int = Field(
        default=50, description="Number of POIs to retrieve from map service"
    )


class MobilityBlockContext(BlockContext):
    next_place: Optional[tuple[str, int]] = Field(
        default=None, description="The next place to go"
    )


class MobilityBlock(Block):
    """
    Main mobility coordination block.
    """

    ParamsType = MobilityBlockParams
    OutputType = SocietyAgentBlockOutput
    ContextType = MobilityBlockContext
    name = "MobilityBlock"
    description = "Used for moving like go to work, go to home, go to other places, etc."
    actions = {
        "place_selection": "Support the place selection action",
        "move": "Support the move action",
        "mobility_none": "Support other mobility operations",
    }

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        block_params: Optional[MobilityBlockParams] = None,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
            block_params=block_params,
        )
        # initialize all blocks
        self.place_selection_block = PlaceSelectionBlock(
            toolbox, agent_memory, self.params.search_limit
        )
        self.move_block = MoveBlock(toolbox, agent_memory)
        self.mobility_none_block = MobilityNoneBlock(toolbox, agent_memory)
        self.trigger_time = 0  # Block invocation counter
        self.token_consumption = 0  # LLM token tracker

        # Initialize block routing system
        self.dispatcher = BlockDispatcher(self._toolbox, agent_memory)
        # register all blocks
        self.dispatcher.register_blocks(
            [self.place_selection_block, self.move_block, self.mobility_none_block]
        )

    async def forward(self, agent_context: DotDict) -> SocietyAgentBlockOutput:
        """Main entry point - delegates to sub-blocks"""
        self.trigger_time += 1
        context = agent_context | self.context
        # Select the appropriate sub-block using dispatcher
        selected_block = await self.dispatcher.dispatch(context)
        if selected_block is None:
            return self.OutputType(
                success=False,
                evaluation=f"Failed to {agent_context['current_step']['intention']}",
                consumed_time=random.randint(1, 30),
                node_id=None,
            )
        # Execute the selected sub-block and get the result
        result = await selected_block.forward(context)  #

        return self.OutputType(**result)



import asyncio
from collections import deque
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Literal, Optional, Union

from fastembed import SparseTextEmbedding

from ..environment import Environment
from ..logger import get_logger
from ..utils.decorators import lock_decorator
from ..vectorstore import VectorStore
from ..agent.memory_config_generator import MemoryConfig

__all__ = [
    "KVMemory",
    "StreamMemory",
    "Memory",
]


class KVMemory:
    def __init__(
        self,
        memory_config: MemoryConfig,
        embedding: SparseTextEmbedding,
    ):
        """
        Initialize the KVMemory with a unified memory configuration.

        - **Args**:
            - `memory_config` (MemoryConfig): The unified memory configuration.
            - `embedding` (SparseTextEmbedding): The embedding object.
        """
        self._memory_config = memory_config
        self._data = {}
        self._vectorstore = VectorStore(embedding)
        self._key_to_doc_id = {}
        self._lock = asyncio.Lock()

        # Initialize from memory config
        for attr in memory_config.attributes.values():
            # Add to init_data
            self._data[attr.name] = deepcopy(attr.default_or_value)

    async def initialize_embeddings(self) -> None:
        """Initialize embeddings for all fields that require them."""
        # Create embeddings for each field that requires it
        documents = []
        keys = []

        # Collect documents and keys from all memory types
        for key, value in self._data.items():
            if self.should_embed(key):
                semantic_text = self._generate_semantic_text(key, value)
                documents.append(semantic_text)
                keys.append(key)

        # Add all documents in one batch
        doc_ids = await self._vectorstore.add_documents(
            documents=documents,
            extra_tags={
                "key": keys,
            },
        )

        # Map document IDs back to their keys
        for key, doc_id in zip(keys, doc_ids):
            self._key_to_doc_id[key] = doc_id

    def _generate_semantic_text(self, key: str, value: Any) -> str:
        """
        Generate semantic text for a given key and value.
        """
        if key in self._memory_config.attributes:
            config = self._memory_config.attributes[key]
            if config.embedding_template:
                return config.embedding_template.format(value)
        return f"My {key} is {value}"

    @lock_decorator
    async def search(
        self, query: str, top_k: int = 3, filter: Optional[dict] = None
    ) -> str:
        """
        Search for relevant memories based on the provided query.

        - **Args**:
            - `query` (str): The text query to search for.
            - `top_k` (int, optional): Number of top relevant memories to return. Defaults to 3.
            - `filter` (Optional[dict], optional): Additional filters for the search. Defaults to None.

        - **Returns**:
            - `str`: Formatted string of the search results.
        """
        filter_dict = {}
        if filter is not None:
            filter_dict.update(filter)
        top_results: list[tuple[str, float, dict]] = (
            await self._vectorstore.similarity_search(
                query=query,
                k=top_k,
                filter=filter_dict,
            )
        )
        # formatted results
        formatted_results = []
        for content, score, metadata in top_results:
            formatted_results.append(f"- {content} ")

        return "Nothing" if len(formatted_results) == 0 else "\n".join(formatted_results)

    def should_embed(self, key: str) -> bool:
        return (
            key in self._memory_config.attributes
            and self._memory_config.attributes[key].whether_embedding
        )

    @lock_decorator
    async def get(
        self,
        key: Any,
        default_value: Optional[Any] = None,
    ) -> Any:
        """
        Retrieve a value from the memory.

        - **Args**:
            - `key` (Any): The key to retrieve.
            - `default_value` (Optional[Any], optional): Default value if the key is not found. Defaults to None.

        - **Returns**:
            - `Any`: The retrieved value or the default value if the key is not found.

        - **Raises**:
            - `KeyError`: If the key is not found in any of the memory sections and no default value is provided.
        """
        if key in self._data:
            return deepcopy(self._data[key])
        else:
            if default_value is None:
                raise KeyError(f"No attribute `{key}` in memories!")
            else:
                return default_value

    @lock_decorator
    async def update(
        self,
        key: Any,
        value: Any,
        mode: Union[Literal["replace"], Literal["merge"]] = "replace",
    ) -> None:
        """
        Update a value in the memory and refresh embeddings if necessary.

        - **Args**:
            - `key` (Any): The key to update.
            - `value` (Any): The new value to set.
            - `mode` (Union[Literal["replace"], Literal["merge"]], optional): Update mode. Defaults to "replace".

        - **Raises**:
            - `ValueError`: If an invalid update mode is provided.
            - `KeyError`: If the key is not found in any of the memory sections.
        """
        # If key doesn't exist, add it directly
        if key not in self._data:
            self._data[key] = value
            # Check if we should embed this field
            if self.should_embed(key):
                semantic_text = self._generate_semantic_text(key, value)
                # Add embedding for new key
                doc_ids = await self._vectorstore.add_documents(
                    documents=[semantic_text],
                    extra_tags={
                        "key": key,
                    },
                )
                self._key_to_doc_id[key] = doc_ids[0]
            return

        # Update existing key
        if mode == "replace":
            # Replace the value directly
            self._data[key] = value

            # Update embeddings if needed
            if self.should_embed(key):
                semantic_text = self._generate_semantic_text(key, value)

                # Delete old embedding if it exists
                if key in self._key_to_doc_id and self._key_to_doc_id[key]:
                    await self._vectorstore.delete_documents(
                        to_delete_ids=[self._key_to_doc_id[key]],
                    )

                # Add new embedding
                doc_ids = await self._vectorstore.add_documents(
                    documents=[semantic_text],
                    extra_tags={
                        "key": key,
                    },
                )
                self._key_to_doc_id[key] = doc_ids[0]

        elif mode == "merge":
            # Get current value
            original_value = self._data[key]

            # Merge based on the type of original value
            if isinstance(original_value, set):
                original_value.update(set(value))
            elif isinstance(original_value, dict):
                original_value.update(dict(value))
            elif isinstance(original_value, list):
                original_value.extend(list(value))
            elif isinstance(original_value, deque):
                original_value.extend(deque(value))
            else:
                # Fall back to replace if merge is not supported
                get_logger().debug(
                    f"Type of {type(original_value)} does not support mode `merge`, using `replace` instead!"
                )
                self._data[key] = value

            # Update embeddings if needed
            if self.should_embed(key):
                semantic_text = self._generate_semantic_text(key, self._data[key])

                # Delete old embedding if it exists
                if key in self._key_to_doc_id and self._key_to_doc_id[key]:
                    await self._vectorstore.delete_documents(
                        to_delete_ids=[self._key_to_doc_id[key]],
                    )

                # Add new embedding
                doc_ids = await self._vectorstore.add_documents(
                    documents=[semantic_text],
                    extra_tags={
                        "key": key,
                    },
                )
                self._key_to_doc_id[key] = doc_ids[0]
        else:
            # Invalid mode
            raise ValueError(f"Invalid update mode `{mode}`!")

    @lock_decorator
    async def export(self, keys: list[str]) -> dict[str, Any]:
        """
        Export the memory of a given keys.
        """
        result = {}
        for k in keys:
            if k in self._data:
                result[k] = deepcopy(self._data[k])
        return result


@dataclass
class MemoryNode:
    """
    A data class representing a memory node.

    - **Attributes**:
        - `topic`: The topic associated with the memory node.
        - `day`: Day of the event or memory.
        - `t`: Time stamp or order.
        - `location`: Location where the event occurred.
        - `description`: Description of the memory.
        - `cognition_id`: ID related to cognitive memory (optional).
        - `id`: Unique ID for this memory node (optional).
    """

    topic: str
    day: int
    t: int
    location: str
    description: str
    cognition_id: Optional[int] = None  # ID related to cognitive memory
    id: Optional[int] = None  # Unique ID for the memory node


class StreamMemory:
    """
    A class used to store and manage time-ordered stream information.

    - **Attributes**:
        - `_memories`: A deque to store memory nodes with a maximum length limit.
        - `_memory_id_counter`: An internal counter to generate unique IDs for each new memory node.
        - `_vectorstore`: The Faiss query object for search functionality.
        - `_status_memory`: The status memory object.
        - `_environment`: The environment object.
    """

    def __init__(
        self,
        environment: Optional[Environment],
        status_memory: KVMemory,
        embedding: SparseTextEmbedding,
        max_len: int = 1000,
    ):
        """
        Initialize an instance of StreamMemory.

        - **Args**:
            - `environment` (Environment): The environment object.
            - `status_memory` (KVMemory): The status memory object.
            - `embedding` (SparseTextEmbedding): The embedding object.
            - `max_len` (int): Maximum length of the deque. Default is 1000.
        """
        self._memories: deque = deque(maxlen=max_len)  # Limit the maximum storage
        self._memory_id_counter: int = 0  # Used for generating unique IDs
        self._status_memory = status_memory
        self._environment = environment
        self._vectorstore = VectorStore(embedding)

    async def add(self, topic: str, description: str) -> int:
        """
        A generic method for adding a memory node and returning the memory node ID.

        - **Args**:
            - `topic` (str): The topic associated with the memory node.
            - `description` (str): Description of the memory.

        - **Returns**:
            - `int`: The unique ID of the newly added memory node.
        """
        if self._environment is None:
            raise ValueError("Environment is not initialized")
        day, t = self._environment.get_datetime()
        position = await self._status_memory.get("position")
        if "aoi_position" in position:
            location = position["aoi_position"]["aoi_id"]
        elif "lane_position" in position:
            location = position["lane_position"]["lane_id"]
        else:
            location = "unknown"

        current_id = self._memory_id_counter
        self._memory_id_counter += 1
        memory_node = MemoryNode(
            topic=topic,
            day=day,
            t=t,
            location=location,
            description=description,
            id=current_id,
        )
        self._memories.append(memory_node)

        # create embedding for new memories
        await self._vectorstore.add_documents(
            documents=[description],
            extra_tags={
                "topic": topic,
                "day": day,
                "time": t,
            },
        )

        return current_id

    async def get_related_cognition(self, memory_id: int) -> Union[MemoryNode, None]:
        """
        Retrieve the related cognition memory node by its ID.

        - **Args**:
            - `memory_id` (int): The ID of the memory to find related cognition for.

        - **Returns**:
            - `Optional[MemoryNode]`: The related cognition memory node, if found; otherwise, None.
        """
        for m in self._memories:
            if m.topic == "cognition" and m.id == memory_id:
                return m
        return None

    async def format_memory(self, memories: list[MemoryNode]) -> str:
        """
        Format a list of memory nodes into a readable string representation.

        - **Args**:
            - `memories` (list[MemoryNode]): List of MemoryNode objects to format.

        - **Returns**:
            - `str`: A formatted string containing the details of each memory node.
        """
        formatted_results = []
        for memory in memories:
            memory_topic = memory.topic
            memory_day = memory.day
            memory_time_seconds = memory.t
            cognition_id = memory.cognition_id

            # Format time
            if memory_time_seconds != "unknown":
                hours = memory_time_seconds // 3600
                minutes = (memory_time_seconds % 3600) // 60
                seconds = memory_time_seconds % 60
                memory_time = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
            else:
                memory_time = "unknown"

            memory_location = memory.location

            # Add cognition information (if exists)
            cognition_info = ""
            if cognition_id is not None:
                cognition_memory = await self.get_related_cognition(cognition_id)
                if cognition_memory:
                    cognition_info = (
                        f"\n  Related cognition: {cognition_memory.description}"
                    )

            formatted_results.append(
                f"- [{memory_topic}]: {memory.description} [day: {memory_day}, time: {memory_time}, "
                f"location: {memory_location}]{cognition_info}"
            )
        return "\n".join(formatted_results)

    async def get_by_ids(self, memory_ids: list[int]) -> str:
        """Retrieve memories by specified IDs"""
        memories = [memory for memory in self._memories if memory.id in memory_ids]
        sorted_results = sorted(memories, key=lambda x: (x.day, x.t), reverse=True)
        return await self.format_memory(sorted_results)

    async def search(
        self,
        query: str,
        topic: Optional[str] = None,
        top_k: int = 3,
        day_range: Optional[tuple[int, int]] = None,
        time_range: Optional[tuple[int, int]] = None,
    ) -> str:
        """
        Search stream memory with optional filters and return formatted results.

        - **Args**:
            - `query` (str): The text to use for searching.
            - `topic` (Optional[str], optional): Filter memories by this topic. Defaults to None.
            - `top_k` (int, optional): Number of top relevant memories to return. Defaults to 3.
            - `day_range` (Optional[tuple[int, int]], optional): Tuple of start and end days for filtering. Defaults to None.
            - `time_range` (Optional[tuple[int, int]], optional): Tuple of start and end times for filtering. Defaults to None.

        - **Returns**:
            - `str`: Formatted string of the search results.
        """
        filter_dict: dict[str, Any] = {"type": "stream"}

        if topic:
            filter_dict["topic"] = topic

        # Add time range filter
        if day_range:
            start_day, end_day = day_range
            filter_dict["day"] = {"gte": start_day, "lte": end_day}

        if time_range:
            start_time, end_time = time_range
            filter_dict["time"] = {"gte": start_time, "lte": end_time}

        top_results = await self._vectorstore.similarity_search(
            query=query,
            k=top_k,
            filter=filter_dict,
        )

        # Sort results by time (first by day, then by time)
        sorted_results = sorted(
            top_results,
            key=lambda x: (x[2].get("day", 0), x[2].get("time", 0)),
            reverse=True,
        )

        formatted_results = []
        for content, score, metadata in sorted_results:
            memory_topic = metadata.get("topic", "unknown")
            memory_day = metadata.get("day", "unknown")
            memory_time_seconds = metadata.get("time", "unknown")
            cognition_id = metadata.get("cognition_id", None)

            # Format time
            if memory_time_seconds != "unknown":
                hours = memory_time_seconds // 3600
                minutes = (memory_time_seconds % 3600) // 60
                seconds = memory_time_seconds % 60
                memory_time = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
            else:
                memory_time = "unknown"

            memory_location = metadata.get("location", "unknown")

            # Add cognition information (if exists)
            cognition_info = ""
            if cognition_id is not None:
                cognition_memory = await self.get_related_cognition(cognition_id)
                if cognition_memory:
                    cognition_info = (
                        f"\n  Related cognition: {cognition_memory.description}"
                    )

            formatted_results.append(
                f"- [{memory_topic}]: {content} [day: {memory_day}, time: {memory_time}, "
                f"location: {memory_location}]{cognition_info}"
            )
        return "Nothing" if len(formatted_results) == 0 else "\n".join(formatted_results)

    async def search_today(
        self,
        query: str = "",  # Optional query text
        topic: Optional[str] = None,
        top_k: int = 100,  # Default to a larger number to ensure all memories of the day are retrieved
    ) -> str:
        """Search all memory events from today

        - **Args**:
            - `query` (`str`): Optional query text, returns all memories of the day if empty. Defaults to "".
            - `topic` (`Optional[str]`): Optional memory topic for filtering specific types of memories. Defaults to None.
            - `top_k` (`int`): Number of most relevant memories to return. Defaults to 100.

        - **Returns**:
            - `str`: Formatted text of today's memories.
        """
        if self._environment is None:
            raise ValueError("Environment is not initialized")
        current_day, _ = self._environment.get_datetime()
        # Use the search method, setting day_range to today
        return await self.search(
            query=query, topic=topic, top_k=top_k, day_range=(current_day, current_day)
        )

    async def add_cognition_to_memory(
        self, memory_ids: list[int], cognition: str
    ) -> None:
        """
        Add cognition to existing memory nodes.

        - **Args**:
            - `memory_ids` (list[int]): List of IDs of the memories to which cognition will be added.
            - `cognition` (str): Description of the cognition to add.
        """
        # Find all corresponding memories
        target_memories = []
        for memory in self._memories:
            if memory.id in memory_ids:
                target_memories.append(memory)

        # Add cognitive memory
        cognition_id = await self.add(topic="cognition", description=cognition)

        # Update the cognition_id of all original memories
        for target_memory in target_memories:
            target_memory.cognition_id = cognition_id

    async def get_all(self) -> list[dict]:
        """
        Retrieve all stream memory nodes as dictionaries.

        - **Returns**:
            - `list[dict]`: List of all memory nodes as dictionaries.
        """
        return [
            {
                "id": memory.id,
                "cognition_id": memory.cognition_id,
                "topic": memory.topic,
                "location": memory.location,
                "description": memory.description,
                "day": memory.day,
                "t": memory.t,
            }
            for memory in self._memories
        ]


class Memory:
    """
    A class to manage different types of memory (status and stream).

    - **Attributes**:
        - `_status` (`KVMemory`): Stores status-related data.
        - `_stream` (`StreamMemory`): Stores stream-related data.
    """

    def __init__(
        self,
        environment: Optional[Environment],
        embedding: SparseTextEmbedding,
        memory_config: MemoryConfig,
    ) -> None:
        """
        Initializes the Memory with a unified memory configuration.

        - **Args**:
            - `environment` (Environment): The environment object.
            - `embedding` (SparseTextEmbedding): The embedding object.
            - `memory_config` (MemoryConfig): The unified memory configuration.
        """
        self._lock = asyncio.Lock()
        self._environment = environment
        self._embedding = embedding

        # Initialize status memory with unified config
        self._status = KVMemory(
            memory_config=memory_config,
            embedding=self._embedding,
        )

        # Add StreamMemory
        self._stream = StreamMemory(
            environment=self._environment,
            embedding=self._embedding,
            status_memory=self._status,
        )

    @property
    def status(self) -> KVMemory:
        return self._status

    @property
    def stream(self) -> StreamMemory:
        return self._stream

    async def initialize_embeddings(self):
        """
        Initialize embeddings within the status memory.

        - **Description**:
            - Asynchronously initializes embeddings for the status memory component, which prepares the system for performing searches.
        """
        await self._status.initialize_embeddings()



from collections import defaultdict
import csv
import io
import json
import logging
import uuid
import zipfile
from typing import List, cast, Dict, Tuple

import yaml
from fastapi import APIRouter, HTTPException, Request, status
from fastapi.responses import StreamingResponse
from sqlalchemy import select, text
from sqlalchemy.ext.asyncio import AsyncSession
from ..models import ApiResponseWrapper
from ..models.agent import (
    agent_dialog,
    agent_profile,
    agent_status,
    agent_survey,
    global_prompt,
)
from ..models.experiment import ApiExperiment, ApiTime, Experiment, ExperimentStatus
from ..models.metric import ApiMetric, metric
from .const import DEMO_USER_ID
from .timezone import ensure_timezone_aware

__all__ = ["router"]

router = APIRouter(tags=["experiments"])


async def _find_started_experiment_by_id(
    request: Request, db: AsyncSession, exp_id: uuid.UUID
) -> Experiment:
    """Find an experiment by ID and check if it has started"""
    tenant_id = await request.app.state.get_tenant_id(request)
    stmt = select(Experiment).where(
        Experiment.tenant_id.in_([tenant_id, "", "default"]), Experiment.id == exp_id
    )
    result = await db.execute(stmt)
    row = result.first()
    if not row:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
        )
    exp: Experiment = row[0]
    if ExperimentStatus(exp.status) == ExperimentStatus.NOT_STARTED:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Experiment not running"
        )
    return exp


@router.get("/experiments")
async def list_experiments(
    request: Request,
) -> ApiResponseWrapper[List[ApiExperiment]]:
    """List all experiments"""
    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = (
            select(Experiment)
            .where(Experiment.tenant_id.in_([tenant_id, "", "default"]))
            .order_by(Experiment.created_at.desc())
        )
        results = await db.execute(stmt)
        db_experiments = [row[0] for row in results.all() if len(row) > 0]

        # 处理时区
        for experiment in db_experiments:
            experiment.created_at = ensure_timezone_aware(experiment.created_at)
            experiment.updated_at = ensure_timezone_aware(experiment.updated_at)

        experiments = cast(List[ApiExperiment], db_experiments)
        return ApiResponseWrapper(data=experiments)


@router.get("/experiments/{exp_id}")
async def get_experiment_by_id(
    request: Request,
    exp_id: uuid.UUID,
) -> ApiResponseWrapper[ApiExperiment]:
    """Get experiment by ID"""

    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = select(Experiment).where(
            Experiment.tenant_id.in_([tenant_id, "", "default"]), Experiment.id == exp_id
        )
        result = await db.execute(stmt)
        row = result.first()
        if not row or len(row) == 0:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
            )
        exp = row[0]
        # 处理时区
        exp.created_at = ensure_timezone_aware(exp.created_at)
        exp.updated_at = ensure_timezone_aware(exp.updated_at)
        return ApiResponseWrapper(data=exp)


@router.get("/experiments/{exp_id}/timeline")
async def get_experiment_status_timeline_by_id(
    request: Request,
    exp_id: uuid.UUID,
) -> ApiResponseWrapper[List[ApiTime]]:
    """Get experiment status timeline by ID"""

    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = select(Experiment).where(
            Experiment.tenant_id.in_([tenant_id, "", "default"]), Experiment.id == exp_id
        )
        result = await db.execute(stmt)
        row = result.first()
        if not row or len(row) == 0:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
            )
        experiment: Experiment = row[0]
        # Check if the experiment has started
        if ExperimentStatus(experiment.status) == ExperimentStatus.NOT_STARTED:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Experiment has not started yet",
            )

        # Get timeline from agent status table
        table_name = experiment.agent_status_tablename

        # the table_name is safe to use in the query
        # it is generated from the experiment id
        query = text(
            f"""
            SELECT day, t 
            FROM {table_name} 
            GROUP BY day, t 
            ORDER BY day, t
        """
        )

        results = (await db.execute(query)).all()
        timeline = [ApiTime(day=int(row[0]), t=float(row[1])) for row in results]

        return ApiResponseWrapper(data=timeline)


@router.delete("/experiments/{exp_id}", status_code=status.HTTP_200_OK)
async def delete_experiment_by_id(
    request: Request,
    exp_id: uuid.UUID,
):
    """Delete experiment by ID"""

    if request.app.state.read_only:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail="Server is in read-only mode"
        )
    tenant_id = await request.app.state.get_tenant_id(request)
    if tenant_id == DEMO_USER_ID:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Demo user is not allowed to delete experiments",
        )

    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        try:
            # Start transaction
            async with db.begin():
                stmt = select(Experiment).where(
                    Experiment.tenant_id == tenant_id, Experiment.id == exp_id
                )
                result = await db.execute(stmt)
                row = result.first()
                if not row or len(row) == 0:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND,
                        detail="Experiment not found",
                    )
                experiment: Experiment = row[0]

                # Get all table names that need to be deleted
                table_names = [
                    experiment.agent_profile_tablename,
                    experiment.agent_status_tablename,
                    experiment.agent_dialog_tablename,
                    experiment.agent_survey_tablename,
                    experiment.global_prompt_tablename,
                    experiment.pending_dialog_tablename,
                    experiment.pending_survey_tablename,
                    experiment.metric_tablename,
                ]

                # Delete related tables
                for table_name in table_names:
                    try:
                        query = text(f"DROP TABLE IF EXISTS {table_name}")
                        await db.execute(query)
                    except Exception as e:
                        logging.error(f"Error dropping table {table_name}: {str(e)}")
                        # Continue execution without interruption

                # Delete experiment record
                await db.delete(experiment)

            # Transaction will be committed automatically

            return ApiResponseWrapper(
                data={"message": "Experiment deleted successfully"}
            )

        except Exception as e:
            logging.error(f"Error deleting experiment: {str(e)}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Failed to delete experiment: {str(e)}",
            )


async def get_experiment_metrics_by_id(
    request: Request,
    db: AsyncSession,
    exp_id: uuid.UUID,
) -> Tuple[bool, Dict[str, List[ApiMetric]]]:
    """Get metrics for an experiment by ID

    Args:
        request: FastAPI request
        db: Database session
        exp_id: Experiment ID

    Returns:
        Tuple containing:
        - bool: Whether metrics were found
        - Dict[str, List[ApiMetric]]: Metrics data aggregated by key
    """

    experiment = await _find_started_experiment_by_id(request, db, exp_id)

    # Get metrics from the metric table
    table_name = experiment.metric_tablename
    
    # Execute query to get metrics data
    query = text(
        f"""
        SELECT key, value, step
        FROM {table_name}
        ORDER BY step
        """
    )
    results = await db.execute(query)
    rows = results.all()

    if not rows:
        return False, {}

    # Aggregate metrics by key
    metrics_by_key: Dict[str, List[ApiMetric]] = defaultdict(list)
    for row in rows:
        api_metric = ApiMetric(
            key=row[0],
            value=float(row[1]),
            step=int(row[2]),
        )
        metrics_by_key[row[0]].append(api_metric)

    return True, metrics_by_key


def serialize_metrics(
    metrics_by_key: Dict[str, List[ApiMetric]],
) -> Dict[str, List[dict]]:
    """Serialize metrics data for JSON output

    Args:
        metrics_by_key: Metrics data aggregated by key

    Returns:
        Dict with serialized metrics data
    """
    return {
        key: [metric.model_dump() for metric in metrics]
        for key, metrics in metrics_by_key.items()
    }


@router.get("/experiments/{exp_id}/metrics")
async def get_experiment_metrics(
    request: Request,
    exp_id: uuid.UUID,
) -> ApiResponseWrapper[Dict[str, List[ApiMetric]]]:
    """Get all metrics for an experiment, aggregated by metric key"""

    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)

        # First verify the experiment exists
        stmt = select(Experiment).where(
            Experiment.tenant_id == tenant_id, Experiment.id == exp_id
        )
        result = await db.execute(stmt)
        row = result.scalar_one_or_none()
        if not row:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
            )

        _, metrics_by_key = await get_experiment_metrics_by_id(request, db, exp_id)
        return ApiResponseWrapper(data=metrics_by_key)


@router.post("/experiments/{exp_id}/export")
async def export_experiment_data(
    request: Request,
    exp_id: uuid.UUID,
) -> StreamingResponse:
    """Export experiment data as a zip file containing YAML and CSV files"""

    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)

        # Get experiment info
        stmt = select(Experiment).where(
            Experiment.tenant_id == tenant_id, Experiment.id == exp_id
        )
        result = await db.execute(stmt)
        row = result.scalar_one_or_none()
        if not row:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
            )
        experiment: Experiment = row

        # Create in-memory zip file
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
            # Export experiment info as YAML
            exp_dict = experiment.to_dict()
            yaml_content = yaml.dump(exp_dict, allow_unicode=True)
            zip_file.writestr("experiment.yaml", yaml_content)

            # Export metrics data as JSON
            found, metrics_by_key = await get_experiment_metrics_by_id(request, db, exp_id)
            if found:
                serialized_metrics = serialize_metrics(metrics_by_key)
                metrics_json = json.dumps(serialized_metrics, indent=2)
                zip_file.writestr("metrics.json", metrics_json)

            # Export artifacts data
            fs_client = request.app.state.env.fs_client
            artifacts_path = f"exps/{tenant_id}/{exp_id}/artifacts.json"
            artifacts_data = fs_client.download(artifacts_path)
            if artifacts_data:
                zip_file.writestr("artifacts.json", artifacts_data)

            # get all tables
            tables = {
                "agent_profile": agent_profile(experiment.agent_profile_tablename),
                "agent_status": agent_status(experiment.agent_status_tablename),
                "agent_survey": agent_survey(experiment.agent_survey_tablename),
                "agent_dialog": agent_dialog(experiment.agent_dialog_tablename),
                "global_prompt": global_prompt(experiment.global_prompt_tablename),
                "metric": metric(experiment.metric_tablename),
            }

            for table_name, (db_table, columns) in tables.items():
                query = select(db_table)
                results = await db.execute(query)
                rows = results.all()

                if rows:
                    # create csv file
                    output = io.StringIO()
                    writer = csv.writer(output)
                    # write header
                    writer.writerow([col for col in columns])
                    # write data
                    writer.writerows(rows)

                    zip_file.writestr(f"{table_name}.csv", output.getvalue())
                    output.close()

        # prepare response
        zip_buffer.seek(0)
        return StreamingResponse(
            iter([zip_buffer.getvalue()]),
            media_type="application/zip",
            headers={
                "Content-Disposition": f"attachment; filename=experiment_{exp_id}_export.zip"
            },
        )


@router.post("/experiments/{exp_id}/artifacts")
async def export_experiment_artifacts(
    request: Request,
    exp_id: uuid.UUID,
) -> StreamingResponse:
    """Export experiment artifacts as a JSON file"""

    tenant_id = await request.app.state.get_tenant_id(request)
    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)

        # Get experiment info
        stmt = select(Experiment).where(
            Experiment.tenant_id == tenant_id, Experiment.id == exp_id
        )
        result = await db.execute(stmt)
        row = result.scalar_one_or_none()
        if not row:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Experiment not found"
            )

        # Get artifacts from S3
        fs_client = request.app.state.env.fs_client
        artifacts_path = f"exps/{tenant_id}/{exp_id}/artifacts.json"
        artifacts_data = fs_client.download(artifacts_path)

        if not artifacts_data:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Artifacts not found"
            )

        return StreamingResponse(
            iter([artifacts_data]),
            media_type="application/json",
            headers={
                "Content-Disposition": f"attachment; filename=experiment_{exp_id}_artifacts.json"
            },
        )



import uuid
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Union,
    cast,
    Any,
    get_args,
    get_origin,
)

from fastapi import APIRouter, Body, HTTPException, Request, status
from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from sqlalchemy import delete, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from ..models import ApiResponseWrapper
from ..models.agent_template import (
    AgentTemplateDB,
    ApiAgentTemplate,
    ChoiceDistributionConfig,
    UniformIntDistributionConfig,
    NormalDistributionConfig,
    AgentParams,
    DistributionType,
)
from .timezone import ensure_timezone_aware

try:
    from agentsociety_community.agents import citizens, supervisors
    from agentsociety_community.workflows import functions as workflow_functions
    from agentsociety_community.blocks import citizens as citizen_blocks
except ImportError:
    import warnings

    warnings.warn(
        "agentsociety_community is not installed. Please install it with `pip install agentsociety-community`"
    )

    citizens = None
    supervisors = None
    workflow_functions = None
    citizen_blocks = None

__all__ = ["router"]

router = APIRouter(tags=["agent_templates"])


@router.get("/agent-templates")
async def list_agent_templates(
    request: Request,
) -> ApiResponseWrapper[List[ApiAgentTemplate]]:
    """List all agent templates"""
    tenant_id = await request.app.state.get_tenant_id(request)

    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = (
            select(AgentTemplateDB)
            .where(AgentTemplateDB.tenant_id.in_([tenant_id, "", "default"]))
            .order_by(AgentTemplateDB.created_at.desc())
        )
        result = await db.execute(stmt)
        templates = result.scalars().all()

        # 处理时区并转换为API格式
        response_templates = []
        for template in templates:
            # 处理时区
            template.created_at = ensure_timezone_aware(template.created_at)
            template.updated_at = ensure_timezone_aware(template.updated_at)
            
            response_templates.append(convert_agent_template_from_db_to_api(template))

        return ApiResponseWrapper(data=response_templates)


@router.get("/agent-templates/{template_id}")
async def get_agent_template(
    request: Request,
    template_id: str,
) -> ApiResponseWrapper[ApiAgentTemplate]:
    """Get agent template by ID"""
    tenant_id = await request.app.state.get_tenant_id(request)

    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = select(AgentTemplateDB).where(
            AgentTemplateDB.tenant_id.in_([tenant_id, "", "default"]),
            AgentTemplateDB.id == template_id,
        )

        result = await db.execute(stmt)
        template = result.scalar_one_or_none()

        if not template:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Template not found"
            )
        api_template = convert_agent_template_from_db_to_api(template)
        return ApiResponseWrapper(data=api_template)


def convert_agent_template_from_db_to_api(
    template: AgentTemplateDB,
) -> ApiAgentTemplate:
    """Convert agent template from DB format to API format"""
    # Parse memory distributions from dictionary format back to Python objects
    memory_distributions_dict = {}

    # Convert each profile into DistributionConfig objects
    for profile_name, profile_config in template.profile.items():
        dist_type = profile_config.get("type")
        if dist_type == DistributionType.CHOICE:
            memory_distributions_dict[profile_name] = ChoiceDistributionConfig(
                type=DistributionType.CHOICE,
                choices=profile_config["choices"],
                weights=profile_config["weights"],
            )
        elif dist_type == DistributionType.UNIFORM_INT:
            memory_distributions_dict[profile_name] = UniformIntDistributionConfig(
                type=DistributionType.UNIFORM_INT,
                min_value=profile_config["min_value"],
                max_value=profile_config["max_value"],
            )
        elif dist_type == DistributionType.NORMAL:
            memory_distributions_dict[profile_name] = NormalDistributionConfig(
                type=DistributionType.NORMAL,
                mean=profile_config["mean"],
                std=profile_config["std"],
            )

    api_template = ApiAgentTemplate(
        tenant_id=template.tenant_id,
        id=template.id,
        name=template.name,
        description=template.description,
        agent_type=template.agent_type,
        agent_class=template.agent_class,
        memory_distributions=memory_distributions_dict,
        agent_params=AgentParams(**template.agent_params),
        blocks=template.blocks,
        created_at=ensure_timezone_aware(template.created_at),
        updated_at=ensure_timezone_aware(template.updated_at),
    )
    return api_template


@router.post("/agent-templates")
async def create_agent_template(
    request: Request,
    template: ApiAgentTemplate = Body(...),
) -> ApiResponseWrapper[ApiAgentTemplate]:
    """Create a new agent template"""
    if request.app.state.read_only:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail="Server is in read-only mode"
        )

    try:
        tenant_id = await request.app.state.get_tenant_id(request)
        template_id = str(uuid.uuid4())

        # Convert memory_distributions to serializable dictionary format
        profile_dict = {}
        for key, value in template.memory_distributions.items():
            # 如果是Config类型，直接转换为字典
            profile_dict[key] = value.model_dump()

        async with request.app.state.get_db() as db:
            db = cast(AsyncSession, db)

            new_template = AgentTemplateDB(
                tenant_id=tenant_id,
                id=template_id,
                name=template.name,
                description=template.description,
                agent_type=template.agent_type,
                agent_class=template.agent_class,
                profile=profile_dict,
                agent_params=template.agent_params.model_dump(),
                blocks=template.blocks,
            )

            db.add(new_template)
            await db.commit()
            await db.refresh(new_template)

            # Construct response data
            response_template = ApiAgentTemplate(
                tenant_id=new_template.tenant_id,
                id=new_template.id,
                name=new_template.name,
                description=new_template.description,
                agent_type=new_template.agent_type,
                agent_class=new_template.agent_class,
                memory_distributions=new_template.profile,
                agent_params=AgentParams(**new_template.agent_params),
                blocks=new_template.blocks,
                created_at=ensure_timezone_aware(new_template.created_at),
                updated_at=ensure_timezone_aware(new_template.updated_at),
            )

            return ApiResponseWrapper(data=response_template)

    except Exception as e:
        print(f"Error details: {str(e)}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Failed to create template: {str(e)}",
        )


@router.put("/agent-templates/{template_id}")
async def update_agent_template(
    request: Request,
    template_id: str,
    template: ApiAgentTemplate = Body(...),
) -> ApiResponseWrapper[ApiAgentTemplate]:
    """Update an existing agent template"""
    if request.app.state.read_only:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail="Server is in read-only mode"
        )

    tenant_id = await request.app.state.get_tenant_id(request)

    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = (
            update(AgentTemplateDB)
            .where(
                AgentTemplateDB.tenant_id == tenant_id,
                AgentTemplateDB.id == template_id,
            )
            .values(
                name=template.name,
                description=template.description,
                agent_type=template.agent_type,
                agent_class=template.agent_class,
                profile=template.memory_distributions,
                agent_params=template.agent_params.model_dump(),
                blocks=template.blocks,
            )
        )

        result = await db.execute(stmt)
        if result.rowcount == 0:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Template not found"
            )

        await db.commit()
        return ApiResponseWrapper(data=template)


@router.delete("/agent-templates/{template_id}")
async def delete_agent_template(
    request: Request,
    template_id: str,
) -> ApiResponseWrapper[Dict[str, str]]:
    """Delete an agent template"""
    if request.app.state.read_only:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail="Server is in read-only mode"
        )

    tenant_id = await request.app.state.get_tenant_id(request)

    async with request.app.state.get_db() as db:
        db = cast(AsyncSession, db)
        stmt = delete(AgentTemplateDB).where(
            AgentTemplateDB.tenant_id == tenant_id, AgentTemplateDB.id == template_id
        )

        result = await db.execute(stmt)
        if result.rowcount == 0:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Template not found"
            )

        await db.commit()
        return ApiResponseWrapper(data={"message": "Template deleted successfully"})


@router.get("/agent-blocks")
async def get_agent_blocks(
    request: Request,
) -> ApiResponseWrapper[List[str]]:
    """Get available block types"""
    try:
        blocks = []

        # 获取citizen blocks
        if citizen_blocks is not None:
            # for block_name, block_class in citizen_blocks.get_type_to_cls_dict().items():
            #     block_info = {
            #         "block_name": block_class.name,
            #         "description": block_class.description,
            #     }
            #     blocks.append(block_info)
            blocks = list(citizen_blocks.get_type_to_cls_dict().keys())

        return ApiResponseWrapper(data=blocks)
    except Exception as e:
        print(f"Error in get_agent_blocks: {str(e)}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Failed to get agent blocks: {str(e)}",
        )


class ParamOption(BaseModel):
    label: str
    value: Any


class Param(BaseModel):
    """
    Parameter edited at frontend
    """

    name: str
    required: bool
    default: Optional[Any] = None
    description: Optional[str] = None
    type: Literal["str", "int", "float", "bool", "select", "select_multiple"]
    options: List[ParamOption] = []


class NameTypeDescription(BaseModel):
    """
    Name, type and description of agent to describe context / status attribute
    """

    name: str
    type: str
    description: Optional[str] = None


def simplify_type(type_annotation):
    """Convert type annotation to a simplified string format for human readable"""
    type_str = str(type_annotation)

    # Handle class format
    if type_str.startswith("<class '") and type_str.endswith("'>"):
        return type_str[8:-2]

    # Handle typing.Optional
    if type_str.startswith("typing.Optional["):
        inner_type = type_str[len("typing.Optional[") : -1]
        return f"{simplify_type(inner_type)}?"

    # Handle typing.List
    if type_str.startswith("typing.List["):
        inner_type = type_str[len("typing.List[") : -1]
        return f"List<{simplify_type(inner_type)}>"

    # Handle typing.Dict
    if type_str.startswith("typing.Dict["):
        # Extract key and value types
        inner_types = type_str[len("typing.Dict[") : -1].split(", ")
        if len(inner_types) == 2:
            key_type = simplify_type(inner_types[0])
            value_type = simplify_type(inner_types[1])
            return f"Dict<{key_type}, {value_type}>"

    # Handle Union type
    if type_str.startswith("typing.Union["):
        inner_types = type_str[len("typing.Union[") : -1].split(", ")
        return " | ".join(simplify_type(t) for t in inner_types)

    return type_str.replace("typing.", "")


def parse_pydantic_to_table(
    model_class: type[BaseModel], exclude_fields: list[str] = []
) -> list[NameTypeDescription]:
    """
    解析pydantic model为适合前端表格呈现的JSON格式

    Args:
        model_class: Pydantic BaseModel子类

    Returns:
        list: 包含表格配置的列表
    """

    if not issubclass(model_class, BaseModel):
        import warnings

        warnings.warn(
            f"model_class {model_class} is not a subclass of BaseModel, return empty list"
        )

    table_items = []
    for field_name, field_info in model_class.model_fields.items():
        if field_name in exclude_fields:
            continue
        table_items.append(
            NameTypeDescription(
                name=field_name,
                type=simplify_type(field_info.annotation),
                description=field_info.description or "",
            )
        )
    return table_items


def parse_pydantic_to_antd_form(
    model_class: type[BaseModel], exclude_fields: list[str] = []
) -> list[Param]:
    """
    解析pydantic model为适合AntDesign表单的JSON格式

    Args:
        model_class: Pydantic BaseModel子类

    Returns:
        list: 包含表格配置的列表
    """

    if not issubclass(model_class, BaseModel):
        import warnings

        warnings.warn(
            f"model_class {model_class} is not a subclass of BaseModel, return empty list"
        )
        return []

    def get_field_type_info(field_annotation, field_info):
        """解析字段类型信息"""
        origin = get_origin(field_annotation)
        args = get_args(field_annotation)

        # 处理Literal类型 - 下拉选择
        if origin is Literal:
            return {
                "type": "select",
                "options": [{"label": str(arg), "value": arg} for arg in args],
            }

        # 处理list类型
        if origin is list:
            if args and get_origin(args[0]) is Literal:
                # list[Literal[...]] - 多选
                literal_args = get_args(args[0])
                return {
                    "type": "select_multiple",
                    "options": [
                        {"label": str(arg), "value": arg} for arg in literal_args
                    ],
                }
            else:
                # TODO: 暂时不支持list[str]类型
                import warnings
                warnings.warn(f"Unsupported field type: {field_annotation}, field_info: {field_info}")
                # raise ValueError(
                #     f"Unsupported field type: {field_annotation}, field_info: {field_info}"
                # )

        # 处理Union类型
        if origin is Union:
            # 简化处理：取第一个非None类型的基本类型（int, float, str, bool, Literal）
            good_types = []
            for arg in args:
                if arg in (int, float, str, bool):
                    good_types.append(arg)
                elif get_origin(arg) in (int, float, str, bool, Literal, Optional):
                    good_types.append(arg)
                elif get_origin(arg) is Union:
                    good_types.extend(get_args(arg))
            print(f"good_types: {good_types}")
            if good_types:
                return get_field_type_info(good_types[0], field_info)
            else:
                raise ValueError(
                    f"Unsupported field type: {field_annotation}, field_info: {field_info}"
                )

        # 处理基础类型
        if field_annotation is str:
            return {"type": "str"}
        elif field_annotation is int:
            return {"type": "int"}
        elif field_annotation is float:
            return {"type": "float"}
        elif field_annotation is bool:
            return {"type": "bool"}

        raise ValueError(
            f"Unsupported field type: {field_annotation}, field_info: {field_info}"
        )

    # 解析模型字段
    form_items = []

    for field_name, field_info in model_class.model_fields.items():
        if field_name in exclude_fields:
            continue

        # 获取字段类型信息
        type_info = get_field_type_info(field_info.annotation, field_info)

        # 构建表单项配置
        form_item = {
            "name": field_name,
            "required": field_info.is_required(),
            "description": field_info.description,
            **type_info,
        }

        # 添加默认值
        if (
            field_info.default is not None
            and field_info.default != ...
            and field_info.default != PydanticUndefined
        ):  # ... 表示无默认值
            form_item["default"] = field_info.default
        elif (
            hasattr(field_info, "default_factory")
            and field_info.default_factory is not None
        ):
            form_item["default"] = field_info.default_factory()  # type: ignore

        form_items.append(form_item)

    params = [Param(**item) for item in form_items]

    return params


class AgentParam(BaseModel):
    """
    Agent parameter
    """

    params_type: List[Param]
    # block_output_type: List[Param]
    context: List[NameTypeDescription]
    status_attributes: List[NameTypeDescription]


@router.get("/agent-param")
async def get_agent_param(
    request: Request,
    agent_type: str,
    agent_class: str,
) -> ApiResponseWrapper[AgentParam]:
    """Get agent's parameters including ParamsType, BlockOutputType, Context and StatusAttributes based on agent type and class"""
    # Get the appropriate agent class based on agent_type and agent_class
    if agent_type == "citizen":
        if citizens is None:
            type_dict = {}
        else:
            type_dict = citizens.get_type_to_cls_dict()
    elif agent_type == "supervisor":
        if supervisors is None:
            type_dict = {}
        else:
            type_dict = supervisors.get_type_to_cls_dict()
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid agent_type. Must be 'citizen' or 'supervisor', got: {agent_type}",
        )

    if agent_class not in type_dict:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid agent_class '{agent_class}' for agent_type '{agent_type}'. Available classes: {list(type_dict.keys())}",
        )

    # Get the agent class
    agent_cls_factory = type_dict[agent_class]
    agent_cls = agent_cls_factory()

    # Get agent parameters information
    param_data = AgentParam(
        params_type=[],
        # block_output_type=[],
        context=[],
        status_attributes=[],
    )

    # Process ParamsType
    param_data.params_type = parse_pydantic_to_antd_form(agent_cls.ParamsType)

    # Process BlockOutputType
    # if agent_cls.BlockOutputType is not None:
    #     param_data["block_output_type"] = parse_pydantic_to_antd_form(
    #         agent_cls.BlockOutputType
    #     )

    # Process Context
    param_data.context = parse_pydantic_to_table(agent_cls.Context)

    # Process StatusAttributes
    param_data.status_attributes = [
        NameTypeDescription(
            name=attr.name,
            type=simplify_type(attr.type),
            # default=(
            #     None
            #     if str(attr.default.__class__).endswith("PydanticUndefinedType'>")
            #     else attr.default
            # ),
            description=attr.description,
            # "whether_embedding": attr.whether_embedding if hasattr(attr, "whether_embedding") else False
        )
        for attr in agent_cls.StatusAttributes
    ]

    return ApiResponseWrapper(data=param_data)

class BlockParam(BaseModel):
    """
    Block parameter
    """

    params_type: List[Param]
    context: List[NameTypeDescription]

@router.get("/block-param/{block_type}")
async def get_block_param(
    request: Request,
    block_type: str,
) -> ApiResponseWrapper[BlockParam]:
    """Get Block's parameters including ParamsType and Context for specified block type"""
    # 在citizen blocks中查找
    if citizen_blocks is not None:
        block_map = citizen_blocks.get_type_to_cls_dict()
        if block_type in block_map:
            block_class_factory = block_map[block_type]
            block_class = block_class_factory()

    if block_class is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid block type: {block_type}",
        )

    # Get Block parameters information
    param_data = BlockParam(
        params_type=[],
        context=[],
    )

    # Process ParamsType
    if hasattr(block_class, "ParamsType"):
        param_data.params_type = parse_pydantic_to_antd_form(
            block_class.ParamsType, exclude_fields=["block_memory"]
        )

    # Process Context
    if hasattr(block_class, "Context"):
        param_data.context = parse_pydantic_to_table(block_class.Context)

    return ApiResponseWrapper(data=param_data)


@router.get("/agent-classes")
async def get_agent_classes(
    request: Request,
    agent_type: str,
) -> ApiResponseWrapper[List[Dict[str, str]]]:
    """Get available agent classes base on agent type"""
    try:
        if agent_type == "citizen":
            if citizens is None:
                type_dict = {}
            else:
                type_dict = citizens.get_type_to_cls_dict()
        elif agent_type == "supervisor":
            if supervisors is None:
                type_dict = {}
            else:
                type_dict = supervisors.get_type_to_cls_dict()
        else:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"Invalid agent_type. Must be 'citizen' or 'supervisor', got: {agent_type}",
            )

        # Convert to list of dicts with value and label for frontend Select component
        agent_type_result = [
            {"value": type_name, "label": type_name} for type_name in type_dict.keys()
        ]
        return ApiResponseWrapper(data=agent_type_result)
    except HTTPException:
        raise
    except Exception as e:
        print(f"Error in get_agent_classes: {str(e)}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Failed to get agent_classes '{agent_type}': {str(e)}",
        )


@router.get("/community/workflow/functions")
async def get_workflow_functions(
    request: Request,
) -> ApiResponseWrapper[List[str]]:
    """Get available workflow function names"""
    try:
        if workflow_functions is None:
            function_map = {}
        else:
            function_map = workflow_functions.get_type_to_cls_dict()
        function_names = list(function_map.keys())
        return ApiResponseWrapper(data=function_names)
    except Exception as e:
        print(f"Error in get_workflow_functions: {str(e)}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Failed to get workflow functions: {str(e)}",
        )




__all__ = [
    "lock_decorator",
]


def lock_decorator(func):
    async def wrapper(self, *args, **kwargs):
        # uid = str(uuid.uuid4())
        # func_name = func.__name__
        # func_module = func.__module__
        async with self._lock:
            # get_logger().info(f"Lock acquired for {uid} with lock id {id(self._lock)}, function: {func_module}.{func_name}")
            result = await func(self, *args, **kwargs)
            # get_logger().info(f"Lock released for {uid} with lock id {id(self._lock)}, function: {func_module}.{func_name}")
            return result

    return wrapper



from datetime import datetime

__all__ = ["ensure_timezone_aware"]


def ensure_timezone_aware(dt: datetime) -> datetime:
    if dt.tzinfo is None:
        # use local timezone instead of UTC
        local_tz = datetime.now().astimezone().tzinfo
        return dt.replace(tzinfo=local_tz)
    return dt



# Due to the current limitations of the simulator's support, only NoneBlock, MessageBlock, and FindPersonBlock are available in the Dispatcher.

from typing import Any, Optional
import json_repair

from ...agent import (
    AgentToolbox,
    Block,
    FormatPrompt,
    BlockParams,
    DotDict,
    BlockContext,
)
from ...logger import get_logger
from ...memory import Memory
from ...agent.dispatcher import BlockDispatcher
from .utils import TIME_ESTIMATE_PROMPT, clean_json_response
from ..sharing_params import SocietyAgentBlockOutput
from pydantic import Field


class MessagePromptManager:
    """
    Manages the creation of message prompts by dynamically formatting templates with agent-specific data.
    """

    def __init__(self):
        pass

    async def get_prompt(
        self, memory, step: dict[str, Any], environment_info: str, target: int, template: str
    ):
        """Generates a formatted prompt for message creation.

        Args:
            memory: Agent's memory to retrieve status data.
            step: Current workflow step containing intention and context.
            target: ID of the target agent for communication.
            template: Raw template string to be formatted.

        Returns:
            Formatted prompt string with placeholders replaced by agent-specific data.
        """
        # Retrieve data
        social_network = await memory.status.get("social_network") or []
        relationship = None
        for relation in social_network:
            if relation.target_id == target:
                relationship = relation
                break
        assert relationship is not None, f"MessagePromptManager: No relation found for target {target}"
        relationship_type = relationship.kind
        relationship_strength = relationship.strength
        chat_histories = await memory.status.get("chat_histories") or {}

        # Build discussion topic constraints
        discussion_constraint = ""
        topics = await memory.status.get("attitude")
        topics = topics.keys()
        if topics:
            topics = ", ".join(f'"{topic}"' for topic in topics)
            discussion_constraint = (
                f"Limit your discussion to the following topics: {topics}."
            )

        # Format prompt
        format_prompt = FormatPrompt(template)
        await format_prompt.format(
            name=await memory.status.get("name", "unknown"),
            gender=await memory.status.get("gender", "unknown"),
            occupation=await memory.status.get("occupation", "unknown"),
            education=await memory.status.get("education", "unknown"),
            personality=await memory.status.get("personality", "unknown"),
            emotion_types=await memory.status.get("emotion_types", "unknown"),
            thought=await memory.status.get("thought", "unknown"),
            background_story=await memory.status.get("background_story", "unknown"),
            relationship_type=relationship_type,
            relationship_strength=relationship_strength,
            chat_history=(
                chat_histories.get(target, "")
                if isinstance(chat_histories, dict)
                else ""
            ),
            intention=step.get("intention", ""),
            discussion_constraint=discussion_constraint,
            environment_info=environment_info,
        )

        return format_prompt.to_dialog()


class SocialNoneBlock(Block):
    """
    NoneBlock
    """

    name = "SocialNoneBlock"
    description = "Handle all other cases if you are not trying to determine the social target or send a message to someone"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(toolbox=toolbox, agent_memory=agent_memory)
        self.guidance_prompt = FormatPrompt(template=TIME_ESTIMATE_PROMPT)

    async def forward(self, context):
        """Executes default behavior when no specific block matches the intention.

        Args:
            step: Current workflow step with 'intention' and other metadata.
            context: Additional execution context (e.g., agent's plan).

        Returns:
            A result dictionary indicating success/failure, time consumed, and execution details.
        """
        intention = str(context["current_step"].get("intention", "socialize"))
        await self.guidance_prompt.format(
            plan=context["plan_context"]["plan"],
            intention=intention,
            emotion_types=await self.memory.status.get("emotion_types"),
        )
        result = await self.llm.atext_request(
            self.guidance_prompt.to_dialog(), response_format={"type": "json_object"}
        )
        result = clean_json_response(result)
        try:
            result: Any = json_repair.loads(result)
            node_id = await self.memory.stream.add(
                topic="social", description=f"I want to: {intention}"
            )
            return {
                "success": True,
                "evaluation": f'Finished {intention}',
                "consumed_time": result["time"],
                "node_id": node_id,
            }
        except Exception as e:
            get_logger().warning(
                f"Error occurred while parsing the evaluation response: {e}, original result: {result}"
            )
            node_id = await self.memory.stream.add(
                topic="social", description=f"I failed to execute {intention}"
            )
            return {
                "success": False,
                "evaluation": f'Failed to execute {intention}',
                "consumed_time": 5,
                "node_id": node_id,
            }


class FindPersonBlock(Block):
    """
    Block for selecting an appropriate agent to socialize with based on relationship strength and context.
    """

    name = "FindPersonBlock"
    description = "Find a suitable person to socialize with"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )

        self.prompt = """
Based on the following information, help me select the most suitable target to interact with:

1. Your Profile:
    - Gender: {gender}
    - Education: {education}
    - Personality: {personality}
    - Occupation: {occupation}
    - Background story: {background_story}

2. Your Current Intention: {intention}

3. Your Current Emotion: {emotion_types}

4. Your Current Thought: {thought}

5. Your social network (shown as id-to-relationship pairs):
    {friend_info}
    Note: For each target, the relationship strength (0-1) indicates how close we are

Please analyze and select:
1. The most appropriate target based on relationship strength and my current intention
2. Whether we should meet online or offline (online: chat, offline: meet in person)

Please output in JSON format, a dictionary:
{{
    "mode": "online" or "offline",
    "target_id": int
}}
        """

    async def forward(self, context: DotDict):
        """Identifies a target agent and interaction mode (online/offline).

        Args:
            context: Additional execution context (may store selected target).

        Returns:
            Result dict with target agent, interaction mode, and execution status.
        """
        try:
            # Get friends list and relationship strength
            my_social_network = await self.memory.status.get("social_network", [])
            if len(my_social_network) == 0:
                node_id = await self.memory.stream.add(
                    topic="social",
                    description="I can't find any target to contact with in my social network.",
                )
                return {
                    "success": False,
                    "evaluation": "No target found in social network.",
                    "consumed_time": 5,
                    "node_id": node_id,
                }

            # Different relationship types
            relationship_info = """
            My social network:
            """
            for relation in my_social_network:
                relationship_info += f"""
                - target_id: {relation.target_id}, relationship_type: {relation.kind}, relationship_strength: {relation.strength}
                """

            # Format the prompt
            formatted_prompt = FormatPrompt(self.prompt)
            await formatted_prompt.format(
                gender=str(await self.memory.status.get("gender")),
                education=str(await self.memory.status.get("education")),
                personality=str(await self.memory.status.get("personality")),
                occupation=str(await self.memory.status.get("occupation")),
                background_story=str(await self.memory.status.get("background_story")),
                intention=str(context["current_step"].get("intention", "socialize")),
                emotion_types=str(await self.memory.status.get("emotion_types")),
                thought=str(await self.memory.status.get("thought")),
                friend_info=relationship_info,
            )

            # Get LLM response
            response = await self.llm.atext_request(
                formatted_prompt.to_dialog(), timeout=300
            )

            try:
                # Parse the response
                response = json_repair.loads(response)
                mode = response["mode"] # type: ignore
                target_id = response["target_id"] # type: ignore

                # Validate the response format
                if not isinstance(mode, str) or mode not in ["online", "offline"]:
                    raise ValueError("Invalid mode")
                if not isinstance(target_id, int):
                    raise ValueError("Invalid target id")

                # Convert index to ID
                target = target_id
                if context is not None:
                    context["target"] = target
            except Exception:
                # If parsing fails, select the friend with the strongest relationship as the default option
                get_logger().warning(f"Error parsing find person response: {response}")
                target = my_social_network[0].target_id
                mode = "online"

            node_id = await self.memory.stream.add(
                topic="social",
                description=f"I selected the friend {target} for {mode} interaction",
            )
            return {
                "success": True,
                "evaluation": f"Selected friend {target} for {mode} interaction",
                "consumed_time": 15,
                "mode": mode,
                "target": target,
                "node_id": node_id,
            }

        except Exception as e:
            get_logger().warning(f"Error in finding person: {e}")
            node_id = await self.memory.stream.add(
                topic="social",
                description="I can't find any friends to socialize with.",
            )
            return {
                "success": False,
                "evaluation": f"Error in finding person: {str(e)}",
                "consumed_time": 5,
                "node_id": node_id,
            }


class MessageBlock(Block):
    """Generate and send messages"""

    name = "MessageBlock"
    description = "Send social message to someone (including online and offline, phone call, social post, etc.)"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.find_person_block = FindPersonBlock(toolbox, agent_memory)

        # configurable fields
        self.default_message_template = """
My name is {name}, I am a {gender}
My occupation is {occupation}. 
My education level is {education}.
My personality is {personality}.
My current emotion is: {emotion_types}.
My current thought is: {thought}.
My background story is: {background_story}.

Now, I want to generate a social message to a target, my relationship with him/her:
Our relationship type is: {relationship_type}
Our relationship strength: {relationship_strength} (0-1, higher is stronger)
My previous chat history with him/her is:
{chat_history}

My intention is: {intention}.

Environment Information:
{environment_info}

Please generate a natural and contextually appropriate message.
Keep it under 100 characters.
The message should reflect my personality and background.

{discussion_constraint}

Please output the message from a first-person perspective, without any other text
"""

        self.prompt_manager = MessagePromptManager()

    async def forward(self, context: DotDict):
        """Generates a message, sends it to the target, and updates chat history.

        Args:
            context: Execution context (may contain pre-selected target).

        Returns:
            Result dict with message content, target, and execution status.
        """
        target = None
        try:
            # Get target from context or find one
            target = context.get("target") if context else None
            if not target:
                result = await self.find_person_block.forward(context)
                if not result["success"]:
                    return {
                        "success": False,
                        "evaluation": "Could not find target for message",
                        "consumed_time": 5,
                        "node_id": result["node_id"],
                    }
                target = result["target"]

            environment_info = """"""
            weather = self.environment.sense("weather")
            if weather:
                environment_info += f"\nCurrent weather: {weather}"
            other_info = self.environment.sense("other_information")
            if other_info:
                environment_info += f"\nOther information: {other_info}"
            if not environment_info:
                environment_info = "No environment information"

            # Get formatted prompt using prompt manager
            formatted_prompt = await self.prompt_manager.get_prompt(
                self.memory,
                context["current_step"],
                environment_info,
                target,
                self.default_message_template,
            )

            # Generate message
            message = await self.llm.atext_request(formatted_prompt, timeout=300)

            # Update chat history with proper format
            chat_histories = await self.memory.status.get("chat_histories") or {}
            if not isinstance(chat_histories, dict):
                chat_histories = {}
            if target not in chat_histories:
                chat_histories[target] = ""
            elif len(chat_histories[target]) > 0:
                chat_histories[target] += ", "
            chat_histories[target] += f"me: {message}"

            await self.memory.status.update("chat_histories", chat_histories)

            # Send message
            await self.agent.send_message_to_agent(target, message, type="social")
            node_id = await self.memory.stream.add(
                topic="social", description=f"I sent a message to {target}: {message}"
            )
            return {
                "success": True,
                "evaluation": f"Sent message to {target}: {message}",
                "consumed_time": 10,
                "node_id": node_id,
            }

        except Exception as e:
            get_logger().warning(f"Error in sending message: {e}")
            node_id = await self.memory.stream.add(
                topic="social", description=f"I can't send a message to {target}"
            )
            return {
                "success": False,
                "evaluation": f"Error in sending message: {str(e)}",
                "consumed_time": 5,
                "node_id": node_id,
            }


class SocialBlockParams(BlockParams): ...


class SocialBlockContext(BlockContext):
    target: Optional[int] = Field(
        default=None,
        description="The target agent id that the agent is going to socialize with",
    )


class SocialBlock(Block):
    """
    Orchestrates social interactions by dispatching to appropriate sub-blocks.
    """

    ParamsType = SocialBlockParams
    OutputType = SocietyAgentBlockOutput
    ContextType = SocialBlockContext
    NeedAgent = True
    name = "SocialBlock"
    description = "Do social interactions, for example, find a friend, send a message, and other social activities."
    actions = {
        "find_person": "Support the find person action, determine the social target.",
        "message": "Support the message action, send a message to the social target.",
        "social_none": "Support other social operations",
    }

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        block_params: Optional[SocialBlockParams] = None,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
            block_params=block_params,
        )
        self.find_person_block = FindPersonBlock(toolbox, agent_memory)
        self.message_block = MessageBlock(toolbox, agent_memory)
        self.noneblock = SocialNoneBlock(toolbox, agent_memory)
        self.dispatcher = BlockDispatcher(toolbox, agent_memory)

        self.trigger_time = 0
        self.token_consumption = 0

        self.dispatcher.register_blocks(
            [self.find_person_block, self.message_block, self.noneblock]
        )

    async def forward(self, agent_context: DotDict) -> SocietyAgentBlockOutput:
        """Main entry point for social interactions. Dispatches to sub-blocks based on context.

        Args:
            step: Workflow step containing intention and metadata.
            context: Additional execution context.

        Returns:
            Result dict from the executed sub-block.
        """
        try:
            self.trigger_time += 1

            context = agent_context | self.context
            self.message_block.set_agent(self.agent)
            # Select the appropriate sub-block using dispatcher
            selected_block = await self.dispatcher.dispatch(context)
            if not selected_block:
                return self.OutputType(
                    success=False,
                    evaluation=f"Failed to complete social interaction with default behavior: {context['current_intention']}",
                    consumed_time=15,
                    node_id=None,
                )
            # Execute the selected sub-block and get the result
            result = await selected_block.forward(context)
            return self.OutputType(**result)

        except Exception as e:
            get_logger().warning(f"Error in social block: {e}")
            return self.OutputType(
                success=False,
                evaluation=f"Failed to complete social interaction with default behavior: {str(e)}",
                consumed_time=15,
                node_id=None,
            )



"""
实用工具
utilities
"""

from .geojson import wrap_feature_collection
from .port import find_free_ports
from .base64 import encode_to_base64

__all__ = [
    "wrap_feature_collection",
    "find_free_ports",
    "encode_to_base64",
]



from pydantic import BaseModel, Field
import boto3
from botocore.exceptions import ClientError

__all__ = ["S3Config"]


class S3Config(BaseModel):
    """S3 configuration"""

    enabled: bool = Field(default=False)
    """Whether S3 is enabled"""

    endpoint: str = ""
    """S3 endpoint"""

    access_key: str = ""
    """S3 access key"""

    secret_key: str = ""
    """S3 secret key"""

    bucket: str = ""
    """S3 bucket"""

    region: str = ""
    """S3 region"""

    prefix: str = Field(default="")
    """S3 prefix"""


class S3Client:
    """S3 client"""

    def __init__(self, config: S3Config):
        """Initialize S3 client"""
        self.config = config

        self.client = boto3.client(
            "s3",
            endpoint_url=config.endpoint,
            aws_access_key_id=config.access_key,
            aws_secret_access_key=config.secret_key,
            region_name=config.region,
            use_ssl=False,
        )

    def _get_s3_path(self, remote_path: str) -> str:
        """Get S3 path"""
        return f"{self.config.prefix.rstrip('/')}/{remote_path.lstrip('/')}"

    def upload(self, data: bytes, remote_path: str):
        """
        Upload bytes data to S3

        Args:
            data: Bytes data to upload
            remote_path: Target path in S3

        Returns:
            bool: Whether the upload is successful
        """
        s3_path = self._get_s3_path(remote_path)
        self.client.put_object(Body=data, Bucket=self.config.bucket, Key=s3_path)

    def download(self, remote_path: str) -> bytes:
        """
        Download bytes data from S3

        Args:
            remote_path: Path of the file in S3

        Returns:
            bytes | None: Downloaded bytes data, None if failed
        """
        s3_path = self._get_s3_path(remote_path)
        response = self.client.get_object(Bucket=self.config.bucket, Key=s3_path)
        return response["Body"].read()

    def exists(self, remote_path: str) -> bool:
        """
        Check if the file exists in S3
        """
        s3_path = self._get_s3_path(remote_path)
        try:
            self.client.head_object(Bucket=self.config.bucket, Key=s3_path)
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise e

    def delete(self, remote_path: str):
        """
        Delete a file from S3
        """
        s3_path = self._get_s3_path(remote_path)
        self.client.delete_object(Bucket=self.config.bucket, Key=s3_path)



import enum
import uuid
from typing import Any, Optional

from pydantic import BaseModel, AwareDatetime

from ...storage import (
    agent_dialog,
    agent_profile,
    agent_status,
    agent_survey,
    global_prompt,
    pending_dialog,
    pending_survey,
)

__all__ = [
    "agent_dialog",
    "agent_profile",
    "agent_status",
    "agent_survey",
    "global_prompt",
    "pending_dialog",
    "pending_survey",
    "AgentDialogType",
    "ApiAgentDialog",
    "ApiAgentProfile",
    "ApiAgentStatus",
    "ApiAgentSurvey",
    "ApiGlobalPrompt",
]

# Database Models


class AgentDialogType(enum.IntEnum):
    """Agent dialog type"""

    Thought = 0  # Dialog in agent self
    Talk = 1  # Dialog with other agents
    User = 2  # Dialog with user


class ApiAgentProfile(BaseModel):
    """Agent profile model for API"""

    id: int
    """Agent ID"""
    name: str
    """Agent name"""
    profile: Any
    """Agent profile (any JSON object)"""

    class Config:
        from_attributes = True


class ApiAgentStatus(BaseModel):
    """Agent status model for API"""

    id: int
    """Agent ID"""
    day: int
    """Day"""
    t: float
    """Time (second)"""
    lng: Optional[float]
    """Longitude"""
    lat: Optional[float]
    """Latitude"""
    parent_id: Optional[int]
    """Parent agent ID"""
    action: str
    """Action"""
    status: Any
    """Status (any JSON object)"""
    created_at: AwareDatetime
    """Created time"""

    class Config:
        from_attributes = True


class ApiAgentSurvey(BaseModel):
    """Agent survey model for API"""

    id: int
    """Agent ID"""
    day: int
    """Day"""
    t: float
    """Time (second)"""
    survey_id: uuid.UUID
    """Survey ID"""
    result: Any
    """Survey result (any JSON object)"""
    created_at: AwareDatetime
    """Created time"""

    class Config:
        from_attributes = True


class ApiAgentDialog(BaseModel):
    """Agent dialog model for API"""

    id: int
    """Agent ID"""
    day: int
    """Day"""
    t: float
    """Time (second)"""
    type: AgentDialogType
    """Dialog type"""
    speaker: str
    """Speaker"""
    content: str
    """Content"""
    created_at: AwareDatetime
    """Created time"""

    class Config:
        from_attributes = True


class ApiGlobalPrompt(BaseModel):
    """Global prompt model for API"""

    day: int
    """Day"""
    t: float
    """Time (second)"""
    prompt: str
    """Prompt"""
    created_at: AwareDatetime
    """Created time"""

    class Config:
        from_attributes = True



from collections.abc import Awaitable
from typing import Any, Literal, TypeVar, Union, overload

from google.protobuf.json_format import MessageToDict, ParseDict
from google.protobuf.message import Message

__all__ = ["parse", "async_parse", "pb2dict", "dict2pb"]

T = TypeVar("T", bound=Message)


@overload
def parse(res: T, dict_return: Literal[True]) -> dict[str, Any]: ...


@overload
def parse(res: T, dict_return: Literal[False]) -> T: ...


def parse(res: T, dict_return: bool) -> Union[dict[str, Any], T]:
    """
    将Protobuf返回值转换为dict或者原始值
    Convert Protobuf return value to dict or original value
    """
    if dict_return:
        return MessageToDict(
            res,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )
    else:
        return res


async def async_parse(res: Awaitable[T], dict_return: bool) -> Union[dict[str, Any], T]:
    """
    将Protobuf await返回值转换为dict或者原始值
    Convert Protobuf await return value to dict or original value
    """
    if dict_return:
        return MessageToDict(
            await res,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )
    else:
        return await res


def pb2dict(pb: Message):
    """
    Convert a protobuf message to a Python dictionary.

    Args:
    - pb: The protobuf message to be converted.

    Returns:
    - The Python dict.
    """
    return MessageToDict(
        pb,
        including_default_value_fields=True,
        preserving_proto_field_name=True,
        use_integers_for_enums=True,
    )


def dict2pb(d: dict, pb: T) -> T:
    """
    Convert a Python dictionary to a protobuf message.

    Args:
    - d: The Python dict to be converted.
    - pb: The protobuf message to be filled.

    Returns:
    - The protobuf message.
    """
    return ParseDict(d, pb, ignore_unknown_fields=True)




from ..utils.grpc import create_aio_channel
from .person_service import PersonService

__all__ = ["CityClient"]


class CityClient:
    """
    模拟器接口
    Simulator interface
    """

    NAME = "city"

    def __init__(
        self,
        url: str,
        secure: bool = False,
    ):
        """
        - **Args**:
            - `url` (`str`): 模拟器server的地址。The address of the emulator server.
            - `secure` (`bool`, `optional`): 是否使用安全连接. Defaults to False. Whether to use a secure connection. Defaults to False.
        """
        aio_channel = create_aio_channel(url, secure)
        self._person_service = PersonService(aio_channel)

    @property
    def person_service(self):
        """
        模拟器智能体服务子模块
        Simulator agent service submodule
        """
        return self._person_service



"""
Commercial Billing System

Provides unified billing functionality for the commercial version.
"""

from typing import Dict, Any

def get_billing_system(config: Dict[str, Any]):
    """Get billing system configuration"""
    if not config.get("enabled", False):
        return None
    
    from .api import router as billing_router
    from .calculator import compute_bill, check_balance, record_experiment_bill
    
    return {
        "router": billing_router,
        "compute_bill": compute_bill,
        "check_balance": check_balance,
        "record_experiment_bill": record_experiment_bill,
        "currency": config.get("currency", "CNY"),
        "rates": config.get("rates", {})
    } 


import uuid
from datetime import datetime
from typing import Any

from pydantic import AwareDatetime, BaseModel
from sqlalchemy.orm import Mapped, mapped_column

from ._base import Base, TABLE_PREFIX

__all__ = ["Survey"]

# Database Models


class Survey(Base):
    """Survey model"""

    __tablename__ = f"{TABLE_PREFIX}survey"

    tenant_id: Mapped[str] = mapped_column(primary_key=True)
    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    name: Mapped[str] = mapped_column()
    data: Mapped[Any] = mapped_column()
    created_at: Mapped[datetime] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime] = mapped_column(
        default=datetime.now, onupdate=datetime.now
    )


# API Models


class ApiSurvey(BaseModel):
    """Survey model for API"""

    id: uuid.UUID
    """Survey ID"""
    name: str
    """Survey name"""
    data: Any
    """Survey data (any JSON object)"""
    created_at: AwareDatetime
    """Created time"""
    updated_at: AwareDatetime
    """Updated time"""

    class Config:
        from_attributes = True



import os
import pickle
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, overload

import numpy as np
import pyproj
import shapely
from google.protobuf.json_format import MessageToDict
from pycityproto.city.map.v2 import map_pb2
from pydantic import BaseModel, Field
from shapely.geometry import Point, Polygon

from ..logger import get_logger
from ..s3 import S3Config, S3Client
from .utils.const import POI_CATG_DICT

__all__ = ["MapData", "MapConfig"]


class MapConfig(BaseModel):
    """Map configuration class."""

    file_path: str = Field(...)
    """Path to the map file. If s3 is enabled, the file will be downloaded from S3"""


class MapData:
    """
    地图API
    Map API
    """

    def __init__(self, config: MapConfig, s3config: S3Config):
        """
        Args:
        - config (MapConfig): Map config, Defaults to None. Map config.
        """
        get_logger().info("MapData init")
        s3client = None
        if s3config.enabled:
            s3client = S3Client(s3config)
        map_data = None
        # 1. try to load from cache
        cache_path = config.file_path+'.cache'
        exists = (
            s3client.exists(cache_path)
            if s3client is not None
            else os.path.exists(cache_path)
        )
        if exists:
            get_logger().info("Start load cache file in MapData")
            if s3client is not None:
                map_bytes = s3client.download(cache_path)
                map_data = pickle.loads(map_bytes)
            else:
                with open(cache_path, "rb") as f:
                    map_data = pickle.load(f)
            get_logger().info("Finish load cache file in MapData")
        if map_data is None:
            get_logger().info("No cache file found, start parse pb file in MapData")
            if s3client is not None:
                map_bytes = s3client.download(config.file_path)
                pb = map_pb2.Map().FromString(map_bytes)
            else:
                with open(config.file_path, "rb") as f:
                    pb = map_pb2.Map().FromString(f.read())

            jsons = []
            # add header
            jsons.append(
                {
                    "class": "header",
                    "data": MessageToDict(
                        pb.header,
                        including_default_value_fields=True,
                        preserving_proto_field_name=True,
                        use_integers_for_enums=True,
                    ),
                }
            )
            # add aois
            for aoi in pb.aois:
                jsons.append(
                    {
                        "class": "aoi",
                        "data": MessageToDict(
                            aoi,
                            including_default_value_fields=True,
                            preserving_proto_field_name=True,
                            use_integers_for_enums=True,
                        ),
                    }
                )
            # add pois
            for poi in pb.pois:
                jsons.append(
                    {
                        "class": "poi",
                        "data": MessageToDict(
                            poi,
                            including_default_value_fields=True,
                            preserving_proto_field_name=True,
                            use_integers_for_enums=True,
                        ),
                    }
                )
            map_data = self._parse_map(jsons)
            get_logger().info("Finish parse pb file")
            if not exists:
                get_logger().info("Start save cache file")
                if s3client is not None:
                    s3client.upload(pickle.dumps(map_data), cache_path)
                else:
                    with open(cache_path, "wb") as f:
                        pickle.dump(map_data, f)
                get_logger().info("Finish save cache file")

        self.header: dict = map_data["header"]
        """
        地图元数据，包含如下属性:
        Map metadata, including the following attributes:
        - name (string): 城市道路名称，供标识数据集合的语义。Map name, to identify the semantics of data collections.
        - date (string): 城市道路数据的创建时间。Map data creation time.
        - north (float): 道路数据的北边界坐标。The coordinate of the northern boundary of the Map data.
        - south (float): 道路数据的南边界坐标。The coordinate of the southern boundary of the Map data.
        - east (float): 道路数据的东边界坐标。The coordinate of the eastern boundary of the Map data.
        - west (float): 道路数据的西边界坐标。The coordinate of the western boundary of the Map data.
        - projection (string): PROJ.4 投影字符串，用以支持xy坐标到其他坐标系的转换。PROJ.4 projection string to support the conversion of xy coordinates to other coordinate systems.
        """

        self.aois: Dict[int, dict] = map_data["aois"]
        """
        地图中的AOI集合（aoi），字典的值包含如下属性:
        AOI collection (aoi) in the map, the value of the dictionary contains the following attributes:
        - id (int): AOI编号。AOI ID.
        - positions (list[XYPosition]): 多边形空间范围。Shape of polygon.
        - area (float): 面积(单位: m2)。Area.
        - driving_positions (list[LanePosition]): 和道路网中行车道的连接点。Connection points to driving lanes.
        - walking_positions (list[LanePosition]): 和道路网中人行道的连接点。Connection points to pedestrian lanes.
        - driving_gates (list[XYPosition]): 和道路网中行车道的连接点对应的AOI边界上的位置。Position on the AOI boundary corresponding to the connection point to driving lanes.
        - walking_gates (list[XYPosition]): 和道路网中人行道的连接点对应的AOI边界上的位置。Position on the AOI boundary corresponding to the connection point to pedestrian lanes.
        - urban_land_use (Optional[str]): 城市建设用地分类，参照执行标准GB 50137-2011（https://www.planning.org.cn/law/uploads/2013/1383993139.pdf） Urban Land use type, refer to the national standard GB 50137-2011.
        - poi_ids (list[int]): 包含的POI列表。Contained POI IDs.
        - shapely_xy (shapely.geometry.Polygon): AOI的形状（xy坐标系）。Shape of polygon (in xy coordinates).
        - shapely_lnglat (shapely.geometry.Polygon): AOI的形状（经纬度坐标系）。Shape of polygon (in latitude and longitude).
        """

        self.pois: Dict[int, dict] = map_data["pois"]
        """
        地图中的POI集合（poi），字典的值包含如下属性:
        POI collection (poi) in the map, the value of the dictionary contains the following attributes:
        - id (int): POI编号。POI ID.
        - name (string): POI名称。POI name.
        - category (string): POI类别编码。POI category code.
        - position (XYPosition): POI位置。POI position.
        - aoi_id (int): POI所属的AOI编号。AOI ID to which the POI belongs.
        """

        (
            self._aoi_tree,
            self._aoi_list,
            self._poi_tree,
            self._poi_list,
        ) = self._build_geo_index()

        self.poi_cate = POI_CATG_DICT

    def get_all_aois(self):
        return self._aoi_list

    def get_aoi(self, aoi_id: int):
        return self.aois[aoi_id]

    def get_all_pois(self):
        return self._poi_list

    def get_poi(self, poi_id: int):
        return self.pois[poi_id]

    def _parse_map(self, m: List[Any]) -> Dict[str, Any]:
        # client = MongoClient(uri)
        # m = list(client[db][coll].find({}))
        get_logger().info("Start parse map data")
        header = None
        aois = {}
        pois = {}
        for d in m:
            if "_id" in d:
                del d["_id"]
            t = d["class"]
            data = d["data"]
            if t == "aoi":
                aois[data["id"]] = data
            elif t == "poi":
                pois[data["id"]] = data
            elif t == "header":
                header = data
        assert header is not None, "header is None"
        get_logger().info("Finish parse map data - classify")
        projector = pyproj.Proj(header["projection"])  #
        # 处理Aoi的Geos
        get_logger().info("Start process aoi geos")
        for aoi in aois.values():
            if "area" not in aoi:
                # 不是多边形aoi
                aoi["shapely_xy"] = Point(
                    aoi["positions"][0]["x"], aoi["positions"][0]["y"]
                )
            else:
                aoi["shapely_xy"] = Polygon(
                    [(one["x"], one["y"]) for one in aoi["positions"]]
                )
            xys = np.array([[one["x"], one["y"]] for one in aoi["positions"]])
            lngs, lats = projector(xys[:, 0], xys[:, 1], inverse=True)
            lnglat_positions = list(zip(lngs, lats))
            if "area" not in aoi:
                aoi["shapely_lnglat"] = Point(lnglat_positions[0])
            else:
                aoi["shapely_lnglat"] = Polygon(lnglat_positions)
        get_logger().info("Finish process aoi geos in MapData")
        # 处理Poi的Geos
        get_logger().info("Start process poi geos in MapData")
        for poi in pois.values():
            point = Point(poi["position"]["x"], poi["position"]["y"])
            poi["shapely_xy"] = point
            lng, lat = projector(point.x, point.y, inverse=True)
            poi["shapely_lnglat"] = Point([lng, lat])
        get_logger().info("Finish process poi geos")

        return {
            "header": header,
            "aois": aois,
            "pois": pois,
        }

    def _build_geo_index(self):
        # poi:
        # {
        #     "id": 700000000,
        #     "name": "天翼(互联网手机卖场)",
        #     "category": "131300",
        #     "position": {
        #       "x": 448802.148620172,
        #       "y": 4412128.118718166
        #     },
        #     "aoi_id": 500018954,
        # }
        get_logger().info("Start build geo index in MapData")
        aoi_list = list(self.aois.values())
        aoi_tree = shapely.STRtree([aoi["shapely_xy"] for aoi in aoi_list])
        poi_list = list(self.pois.values())
        poi_tree = shapely.STRtree([poi["shapely_xy"] for poi in poi_list])
        get_logger().info("Finish build geo index in MapData")
        return (
            aoi_tree,
            aoi_list,
            poi_tree,
            poi_list,
        )

    @overload
    def query_pois(
        self,
        center: Union[Tuple[float, float], Point],
        radius: Optional[float] = None,
        category_prefix: Optional[str] = None,
        limit: Optional[int] = None,
        return_distance: Literal[False] = False,
    ) -> List[Any]: ...

    @overload
    def query_pois(
        self,
        center: Union[Tuple[float, float], Point],
        radius: Optional[float] = None,
        category_prefix: Optional[str] = None,
        limit: Optional[int] = None,
        return_distance: Literal[True] = True,
    ) -> List[Tuple[Any, float]]: ...

    def query_pois(
        self,
        center: Union[Tuple[float, float], Point],
        radius: Optional[float] = None,
        category_prefix: Optional[str] = None,
        limit: Optional[int] = None,
        return_distance: bool = True,
    ) -> Union[List[Tuple[Any, float]], List[Any]]:
        """
        查询center点指定半径内类别满足前缀的poi（按距离排序）。Query the POIs whose categories satisfy the prefix within the specified radius of the center point (sorted by distance).

        Args:
        - center (x, y): 中心点（xy坐标系）。Center point (xy coordinate system).
        - radius (float, optional): 半径（单位：m）。如果不提供则返回所有的poi。Radius (unit: m).If not provided, all pois within the map will be returned.
        - category_prefix (str, optional): 类别前缀，如实际类别为100000，那么匹配的前缀可以为10、1000等。Category prefix, if the actual category is 100000, then the matching prefix can be 10, 1000, etc.
        - limit (int, optional): 最多返回的poi数量，按距离排序，近的优先（默认None）. The maximum number of POIs returned, sorted by distance, closest ones first (default to None).
        - return_distance (bool): 是否返回距离。Return the distance or not.

        Returns:
        - Union[List[Tuple[Any, float]],List[Any]]: poi列表，每个元素为（poi, 距离）或者poi。poi list, each element is (poi, distance) or poi.
        """
        if not isinstance(center, Point):
            center = Point(center)
        if radius is None:
            if return_distance:
                pois = [(p, center.distance(p["shapely_xy"])) for p in self._poi_list]
            else:
                pois = [p for p in self._poi_list]
        else:
            # 获取半径内的poi
            indices = self._poi_tree.query(center.buffer(radius))
            # 过滤掉不满足类别前缀的poi
            pois = []
            for index in indices:
                poi = self._poi_list[index]
                if category_prefix is None or poi["category"].startswith(
                    category_prefix
                ):
                    if return_distance:
                        distance = center.distance(poi["shapely_xy"])
                        pois.append((poi, distance))
                    else:
                        pois.append(poi)
        if return_distance:
            # 按照距离排序
            pois = sorted(pois, key=lambda x: x[1])
        if limit is not None:
            pois = pois[:limit]
        return pois

    def get_poi_cate(self):
        return self.poi_cate

    def get_map_header(self):
        return self.header

    def get_projector(self):
        return self.header["projection"]

    def query_aois(
        self,
        center: Union[Tuple[float, float], Point],
        radius: float,
        urban_land_uses: Optional[List[str]] = None,
        limit: Optional[int] = None,
    ) -> List[Tuple[Any, float]]:
        """
        查询center点指定半径内城市用地满足条件的aoi（按距离排序）。Query the AOIs whose urban land use within the specified radius of the center point meets the conditions (sorted by distance).

        Args:
        - center (x, y): 中心点（xy坐标系）。Center point (xy coordinate system).
        - radius (float): 半径（单位：m）。Radius (unit: m).
        - urban_land_uses (List[str], optional): 城市用地分类列表，参照执行标准GB 50137-2011（https://www.planning.org.cn/law/uploads/2013/1383993139.pdf）. Urban land use classification list, refer to the national standard GB 50137-2011.
        - limit (int, optional): 最多返回的aoi数量，按距离排序，近的优先（默认None）. The maximum number of AOIs returned, sorted by distance, closest ones first (default to None).

        Returns:
        - List[Tuple[Any, float]]: aoi列表，每个元素为（aoi, 距离）。aoi list, each element is (aoi, distance).
        """

        if not isinstance(center, Point):
            center = Point(center)
        # 获取半径内的aoi
        indices = self._aoi_tree.query(center.buffer(radius))
        # 过滤掉不满足城市用地条件的aoi
        aois = []
        for index in indices:
            aoi = self._aoi_list[index]
            if (
                urban_land_uses is not None
                and aoi["urban_land_use"] not in urban_land_uses
            ):
                continue
            distance = center.distance(aoi["shapely_xy"])
            aois.append((aoi, distance))
        # 按照距离排序
        aois = sorted(aois, key=lambda x: x[1])
        if limit is not None:
            aois = aois[:limit]
        return aois



from typing import Any

import json_repair

from ...agent import AgentToolbox, Block, FormatPrompt, DotDict
from ...logger import get_logger
from ...memory import Memory
from .utils import clean_json_response

INITIAL_NEEDS_PROMPT = """You are an intelligent agent satisfaction initialization system. Based on the profile information below, please help initialize the agent's satisfaction levels and related parameters.

Profile Information:
- Gender: ${profile.gender}
- Education Level: ${profile.education} 
- Consumption Level: ${profile.consumption}
- Occupation: ${profile.occupation}
- Age: ${profile.age}
- Monthly Income: ${profile.income}

Current Time: ${context.current_time}

Please initialize the agent's satisfaction levels and parameters based on the profile above. Return the values in JSON format with the following structure:

Current satisfaction levels (0-1 float values, lower means less satisfied):
- hunger_satisfaction: Hunger satisfaction level (Normally, the agent will be less satisfied with hunger at eating time)
- energy_satisfaction: Energy satisfaction level (Normally, at night, the agent will be less satisfied with energy)
- safety_satisfaction: Safety satisfaction level (Normally, the agent will be more satisfied with safety when they have high income and currency)
- social_satisfaction: Social satisfaction level

Please response in json format, example:
{{
    "current_satisfaction": {{
        "hunger_satisfaction": 0.8,
        "energy_satisfaction": 0.7,
        "safety_satisfaction": 0.9,
        "social_satisfaction": 0.6
    }}
}}
DO NOT INCLUDE ANY COMMENTS IN YOUR RESPONSE.
DO NOT INCLUDE ANY COMMENTS IN YOUR RESPONSE.
DO NOT INCLUDE ANY COMMENTS IN YOUR RESPONSE.
"""

EVALUATION_PROMPT = """You are an evaluation system for an intelligent agent. The agent has performed the following actions to satisfy the {current_need} need:

Goal: {plan_target}
Execution situation:
{evaluation_results}

Current satisfaction: 
- hunger_satisfaction: {hunger_satisfaction}
- energy_satisfaction: {energy_satisfaction}
- safety_satisfaction: {safety_satisfaction}
- social_satisfaction: {social_satisfaction}

Please evaluate and adjust the value of {current_need} satisfaction based on the execution results above.

Notes:
1. Satisfaction values range from 0-1, where:
   - 1 means the need is fully satisfied
   - 0 means the need is completely unsatisfied 
   - Higher values indicate greater need satisfaction
2. If the current need is not "whatever", only return the new value for the current need. Otherwise, return both safe and social need values.
3. Ensure the return value is in valid JSON format, examples below:

Please response in json format for specific need (hungry here) adjustment (Do not return any other text), example:
{{
    "hunger_satisfaction": new_hunger_satisfaction_value
}}

Please response in json format for whatever need adjustment (Do not return any other text), example:
{{
    "safety_satisfaction": new_safety_satisfaction_value,
    "social_satisfaction": new_social_satisfaction_value
}}
"""

REFLECTION_PROMPT = """You are an intelligent agent reflection system. Based on the intervention message below, please help to rebuild the satisfaction levels of the agent.

The agent has received/sense the following intervention message:
--------------------------------
{intervention_message}
--------------------------------

And the agent's current needs are:
- hunger_satisfaction: {hunger_satisfaction}
- energy_satisfaction: {energy_satisfaction}
- safety_satisfaction: {safety_satisfaction}
- social_satisfaction: {social_satisfaction}

The agent's current action is:
--------------------------------
{current_action}
--------------------------------

Please response in json format, example:
{{
    "hunger_satisfaction": new_hunger_satisfaction_value,
    "energy_satisfaction": new_energy_satisfaction_value,
    "safety_satisfaction": new_safety_satisfaction_value,
    "social_satisfaction": new_social_satisfaction_value,
}}
If you think the agent has to stop the current action and do something to satisfy the needs, please response in json format, example:
{{
    "do_something": True,
    "description": "Go to the hospital"
}}
"""


class NeedsBlock(Block):
    """
    Manages agent's dynamic needs system including:
    - Initializing satisfaction levels
    - Time-based decay of satisfaction values
    - Need prioritization based on thresholds
    - Plan execution evaluation and satisfaction adjustments
    """

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        agent_context: DotDict,
        evaluation_prompt: str = EVALUATION_PROMPT,
        reflection_prompt: str = REFLECTION_PROMPT,
        initial_prompt: str = INITIAL_NEEDS_PROMPT,
    ):
        """
        Initialize needs management system.

        Args:
            llm: Language model instance for processing prompts
            environment: Simulation environment controller
            agent_memory: Agent's memory storage interface

        Configuration Parameters:
            alpha_H: Hunger satisfaction decay rate per hour (default: 0.15)
            alpha_D: Energy satisfaction decay rate per hour (default: 0.08)
            alpha_P: Safety satisfaction decay rate per hour (default: 0.05)
            alpha_C: Social satisfaction decay rate per hour (default: 0.1)
            T_H: Hunger threshold for triggering need (default: 0.2)
            T_D: Energy threshold for triggering need (default: 0.2)
            T_P: Safety threshold for triggering need (default: 0.2)
            T_C: Social threshold for triggering need (default: 0.3)
        """
        super().__init__(toolbox=toolbox, agent_memory=agent_memory)
        self.context = agent_context
        self.evaluation_prompt = FormatPrompt(template=evaluation_prompt)
        self.initial_prompt = FormatPrompt(template=initial_prompt, memory=agent_memory)
        self.reflection_prompt = FormatPrompt(template=reflection_prompt)
        self.need_work = True
        self.now_day = -1
        self.last_evaluation_time = None
        self.trigger_time = 0
        self.token_consumption = 0
        self.initialized = False
        self.alpha_H, self.alpha_D, self.alpha_P, self.alpha_C = (
            0.15,
            0.08,
            0.05,
            0.1,
        )  # Hunger decay rate, Energy decay rate, Safety decay rate, Social decay rate
        self.T_H, self.T_D, self.T_P, self.T_C = (
            0.2,
            0.2,
            0.2,
            0.3,
        )  # Hunger threshold, Energy threshold, Safety threshold, Social threshold
        # intervention need
        self._need_to_do = None
        # determine if the intervention need has been checked
        self._need_to_do_checked = False

    async def reset(self):
        """Reset the needs block."""
        self._need_to_do = None
        self._need_to_do_checked = False
        self.initialized = False

    async def initialize(self):
        """
        Initialize agent's satisfaction levels using profile data.
        - Runs once per simulation day
        - Collects demographic data from memory
        - Generates initial satisfaction values via LLM
        - Handles JSON parsing and validation
        """
        day, t = self.environment.get_datetime()
        if day != self.now_day and t >= 7 * 60 * 60:
            self.now_day = day
            workday = self.environment.sense("workday")
            if workday:
                self.need_work = True
            else:
                self.need_work = False

        if not self.initialized:
            await self.initial_prompt.format(context=self.context)
            response = await self.llm.atext_request(
                self.initial_prompt.to_dialog(), response_format={"type": "json_object"}
            )
            response = clean_json_response(response)
            retry = 3
            while retry > 0:
                try:
                    satisfaction: Any = json_repair.loads(response)
                    satisfactions = satisfaction["current_satisfaction"]
                    await self.memory.status.update(
                        "hunger_satisfaction", satisfactions["hunger_satisfaction"]
                    )
                    await self.memory.status.update(
                        "energy_satisfaction", satisfactions["energy_satisfaction"]
                    )
                    await self.memory.status.update(
                        "safety_satisfaction", satisfactions["safety_satisfaction"]
                    )
                    await self.memory.status.update(
                        "social_satisfaction", satisfactions["social_satisfaction"]
                    )
                    break
                except Exception as e:
                    get_logger().warning(f"Initial response error: {e}")
                    retry -= 1

            current_plan = await self.memory.status.get("current_plan", False)
            if current_plan:
                history = await self.memory.status.get("plan_history")
                history.append(current_plan)
                await self.memory.status.update("plan_history", history)
                await self.memory.status.update("current_plan", None)
                await self.memory.status.update("execution_context", {})
            self.initialized = True

    async def reflect_to_intervention(self, intervention: str):
        # rebuild needs for intervention
        current_plan = await self.memory.status.get("current_plan", False)
        if not current_plan:
            return
        step_index = current_plan.get("index", 0)
        current_action = current_plan.get("steps", [{"intention": "", "type": ""}])[
            step_index
        ]
        action_message = (
            f"{current_action['intention']} ({current_action['type']})"
            if current_action["intention"] != ""
            else "None"
        )
        await self.reflection_prompt.format(
            intervention_message=intervention,
            current_action=action_message,
            hunger_satisfaction=await self.memory.status.get("hunger_satisfaction"),
            energy_satisfaction=await self.memory.status.get("energy_satisfaction"),
            safety_satisfaction=await self.memory.status.get("safety_satisfaction"),
            social_satisfaction=await self.memory.status.get("social_satisfaction"),
        )
        response = await self.llm.atext_request(
            self.reflection_prompt.to_dialog(), response_format={"type": "json_object"}
        )
        try:
            reflection: Any = json_repair.loads(clean_json_response(response))
            if "do_something" in reflection:
                self._need_to_do = reflection["description"]
            else:
                # update satisfaction
                for need_type, new_value in reflection.items():
                    if need_type in [
                        "hunger_satisfaction",
                        "energy_satisfaction",
                        "safety_satisfaction",
                        "social_satisfaction",
                    ]:
                        await self.memory.status.update(need_type, new_value)
        except Exception as e:
            get_logger().warning(f"Error processing reflection response: {str(e)}")
            get_logger().warning(f"Original response: {response}")
            return None

    async def time_decay(self):
        """
        Apply time-based decay to satisfaction values.
        - Calculates hours since last update
        - Applies exponential decay to each satisfaction dimension
        - Ensures values stay within [0,1] range
        """
        # calculate time diff
        tick_now = self.environment.get_tick()
        if self.last_evaluation_time is None:
            self.last_evaluation_time = tick_now
            return
        else:
            time_diff = (tick_now - self.last_evaluation_time) / 3600
            self.last_evaluation_time = tick_now

        # acquire current satisfaction
        hunger_satisfaction = await self.memory.status.get("hunger_satisfaction")
        energy_satisfaction = await self.memory.status.get("energy_satisfaction")
        safety_satisfaction = await self.memory.status.get("safety_satisfaction")
        social_satisfaction = await self.memory.status.get("social_satisfaction")

        # calculates hunger and fatigue decay based on elapsed time
        hungry_decay = self.alpha_H * time_diff
        energy_decay = self.alpha_D * time_diff
        safety_decay = self.alpha_P * time_diff
        social_decay = self.alpha_C * time_diff
        hunger_satisfaction = max(0, hunger_satisfaction - hungry_decay)
        energy_satisfaction = max(0, energy_satisfaction - energy_decay)
        safety_satisfaction = max(0, safety_satisfaction - safety_decay)
        social_satisfaction = max(0, social_satisfaction - social_decay)

        # update satisfaction
        await self.memory.status.update("hunger_satisfaction", hunger_satisfaction)
        await self.memory.status.update("energy_satisfaction", energy_satisfaction)
        await self.memory.status.update("safety_satisfaction", safety_satisfaction)
        await self.memory.status.update("social_satisfaction", social_satisfaction)

    async def update_when_plan_completed(self):
        # Check if there is any ongoing plan
        current_plan = await self.memory.status.get("current_plan")
        if current_plan and (
            current_plan.get("completed") or current_plan.get("failed")
        ):
            # Evaluate the execution process of the plan and adjust needs
            pre_need = await self.memory.status.get("current_need")
            # evaluate plan execution and adjust needs
            await self.evaluate_and_adjust_needs(current_plan)
            # add completed plan to history
            history = await self.memory.status.get("plan_history")
            history.append(current_plan)
            await self.memory.status.update("plan_history", history)
            await self.memory.status.update("current_plan", None)
            await self.memory.status.update("execution_context", {})
            if pre_need == self._need_to_do:
                self._need_to_do = None
                self._need_to_do_checked = False

    async def determine_current_need(self):
        """
        Determine agent's current dominant need based on:
        - Satisfaction thresholds
        - Need priority hierarchy (hungry > tired > safe > social)
        - Workday requirements
        - Ongoing plan interruptions
        """
        cognition = None
        hunger_satisfaction = await self.memory.status.get("hunger_satisfaction")
        energy_satisfaction = await self.memory.status.get("energy_satisfaction")
        safety_satisfaction = await self.memory.status.get("safety_satisfaction")
        social_satisfaction = await self.memory.status.get("social_satisfaction")

        # If needs adjustment is required, update current need
        # The adjustment scheme is to adjust the need if the current need is empty, or a higher priority need appears
        current_plan = await self.memory.status.get("current_plan")
        current_need = await self.memory.status.get("current_need")

        # When there's no plan, get all satisfaction values and check each need against its threshold based on priority
        if not current_plan:
            # check needs in priority order
            if self._need_to_do:
                await self.memory.status.update("current_need", self._need_to_do)
                self.context.current_intention = self._need_to_do
                await self.memory.stream.add(
                    topic="cognition", description=f"I need to do: {self._need_to_do}"
                )
                cognition = f"I need to do: {self._need_to_do}"
                self._need_to_do_checked = True
            elif hunger_satisfaction <= self.T_H:
                await self.memory.status.update("current_need", "hungry")
                self.context.current_intention = "hungry"
                await self.memory.stream.add(
                    topic="cognition", description="I feel hungry"
                )
                cognition = "I feel hungry"
            elif energy_satisfaction <= self.T_D:
                await self.memory.status.update("current_need", "tired")
                self.context.current_intention = "tired"
                await self.memory.stream.add(
                    topic="cognition", description="I feel tired"
                )
                cognition = "I feel tired"
            elif self.need_work:
                await self.memory.status.update("current_need", "safe")
                self.context.current_intention = "safe"
                await self.memory.stream.add(
                    topic="cognition", description="I need to work"
                )
                cognition = "I need to work"
                self.need_work = False
            elif safety_satisfaction <= self.T_P:
                await self.memory.status.update("current_need", "safe")
                self.context.current_intention = "safe"
                await self.memory.stream.add(
                    topic="cognition", description="I have safe needs right now"
                )
                cognition = "I have safe needs right now"
            elif social_satisfaction <= self.T_C:
                await self.memory.status.update("current_need", "social")
                self.context.current_intention = "social"
                await self.memory.stream.add(
                    topic="cognition", description="I have social needs right now"
                )
                cognition = "I have social needs right now"
            else:
                await self.memory.status.update("current_need", "whatever")
                self.context.current_intention = "whatever"
                await self.memory.stream.add(
                    topic="cognition", description="I have no specific needs right now"
                )
                cognition = "I have no specific needs right now"

        else:
            # While there is an ongoing plan, only adjust for higher priority needs
            needs_changed = False
            new_need = None
            if self._need_to_do:
                if not self._need_to_do_checked:
                    new_need = self._need_to_do
                    needs_changed = True
                    self._need_to_do_checked = True
                    cognition = f"I need to change my plan, as {self._need_to_do} is more important than {current_need}"
                else:
                    cognition = f"I still need to concentrate on {self._need_to_do}"
                    pass  # still concentrate on the emergency need
            elif hunger_satisfaction <= self.T_H and current_need not in [
                "hungry",
                "tired",
            ]:
                new_need = "hungry"
                needs_changed = True
            elif energy_satisfaction <= self.T_D and current_need not in [
                "hungry",
                "tired",
            ]:
                new_need = "tired"
                needs_changed = True
            elif safety_satisfaction <= self.T_P and current_need not in [
                "hungry",
                "tired",
                "safe",
            ]:
                new_need = "safe"
                needs_changed = True
            elif social_satisfaction <= self.T_C and current_need not in [
                "hungry",
                "tired",
                "safe",
                "social",
            ]:
                new_need = "social"
                needs_changed = True

            # If needs have changed, interrupt the current plan
            if needs_changed:
                await self.evaluate_and_adjust_needs(current_plan)
                history = await self.memory.status.get("plan_history")
                history.append(current_plan)
                await self.memory.stream.add(
                    topic="cognition",
                    description=f"I need to change my plan because the need of [{new_need}] is more important than [{current_need}]",
                )
                cognition = f"I need to change my plan because the need of [{new_need}] is more important than [{current_need}]"
                await self.memory.status.update("current_need", new_need)
                self.context.current_intention = new_need
                await self.memory.status.update("plan_history", history)
                await self.memory.status.update("current_plan", None)
                await self.memory.status.update("execution_context", {})
        return cognition

    async def evaluate_and_adjust_needs(self, completed_plan):
        """
        Evaluate plan execution results and adjust satisfaction values.
        - Extracts step evaluations from completed plan
        - Constructs evaluation prompt for LLM
        - Processes LLM response and updates satisfaction values
        - Implements retry logic for invalid responses
        """
        # Retrieve the executed plan and evaluation results
        evaluation_results = []
        for step in completed_plan["steps"]:
            if "evaluation" in step["evaluation"]:
                eva_ = step["evaluation"]["evaluation"]
            else:
                eva_ = "Plan failed or skipped, not completed"
            evaluation_results.append(f"- {step['intention']} ({step['type']}): {eva_}")
        evaluation_results = "\n".join(evaluation_results)

        # Use LLM for evaluation
        current_need = await self.memory.status.get("current_need")
        await self.evaluation_prompt.format(
            current_need=current_need,
            plan_target=completed_plan["target"],
            evaluation_results=evaluation_results,
            hunger_satisfaction=await self.memory.status.get("hunger_satisfaction"),
            energy_satisfaction=await self.memory.status.get("energy_satisfaction"),
            safety_satisfaction=await self.memory.status.get("safety_satisfaction"),
            social_satisfaction=await self.memory.status.get("social_satisfaction"),
        )

        retry = 3
        while retry > 0:
            response = await self.llm.atext_request(
                self.evaluation_prompt.to_dialog(),
                response_format={"type": "json_object"},
            )
            try:
                new_satisfaction: Any = json_repair.loads(clean_json_response(response))  # type: ignore
                # Update values of all needs
                for need_type, new_value in new_satisfaction.items():
                    if need_type in [
                        "hunger_satisfaction",
                        "energy_satisfaction",
                        "safety_satisfaction",
                        "social_satisfaction",
                    ]:
                        await self.memory.status.update(need_type, new_value)
                return
            except Exception as e:
                get_logger().warning(f"Error processing evaluation response: {str(e)}")
                get_logger().warning(f"Original response: {response}")
                retry -= 1

    async def forward(self):
        """
        Main execution flow for needs management:
        1. Initialize satisfaction values (if needed)
        2. Apply time-based decay
        3. Handle completed plans
        4. Determine current dominant need
        """
        cognition = None

        await self.initialize()

        # satisfaction decay with time
        await self.time_decay()

        # update when plan completed
        await self.update_when_plan_completed()

        # determine current need
        cognition = await self.determine_current_need()

        return cognition



import asyncio
from concurrent import futures
from typing import Dict, Optional

import grpc
from grpc.aio import Server as AioServer
from grpc.aio import ServicerContext as AioServicerContext
from pycityproto.city.sync.v2 import sync_service_pb2 as sync_service
from pycityproto.city.sync.v2 import sync_service_pb2_grpc as sync_grpc

from ...logger import get_logger

__all__ = ["Syncer"]


class Syncer(sync_grpc.SyncServiceServicer):
    def __init__(self, addr: str):
        """
        - **Args**:
            - `addr` (str): The address of the syncer
        """
        self.addr = addr

        self._enter_barrier = asyncio.Barrier(2)
        self._exit_barrier = asyncio.Barrier(2)

        # server
        self.server: Optional[AioServer] = None

        self._name2url: Dict[str, str] = {}

    async def init(self) -> None:
        # Create gRPC server
        self.server = grpc.aio.server(futures.ThreadPoolExecutor(max_workers=1))
        sync_grpc.add_SyncServiceServicer_to_server(self, self.server)
        self.server.add_insecure_port(self.addr)

        await self.server.start()
        get_logger().info(f"server listening at {self.addr}")

    async def close(self) -> None:
        if self.server:
            await self.server.stop(1)
            await self.server.wait_for_termination()
            self.server = None

    def __del__(self):
        if self.server:
            # The final mechanism for shutting down the server to avoid a situation where it is impossible to exit
            loop = asyncio.new_event_loop()
            loop.run_until_complete(self.close())
            loop.close()

    async def SetURL(
        self, request: sync_service.SetURLRequest, context: AioServicerContext
    ) -> sync_service.SetURLResponse:
        """Set the URL of the syncer"""
        self._name2url[request.name] = request.url
        return sync_service.SetURLResponse()

    async def GetURL(
        self, request: sync_service.GetURLRequest, context: AioServicerContext
    ) -> sync_service.GetURLResponse:
        """Get the URL of the syncer"""
        return sync_service.GetURLResponse(url=self._name2url[request.name])

    async def EnterStepSync(
        self, request: sync_service.EnterStepSyncRequest, context: AioServicerContext
    ) -> sync_service.EnterStepSyncResponse:
        get_logger().debug(f"EnterStepSync: {request.name}")
        await self._enter_barrier.wait()
        return sync_service.EnterStepSyncResponse()

    async def ExitStepSync(
        self, request: sync_service.ExitStepSyncRequest, context: AioServicerContext
    ) -> sync_service.ExitStepSyncResponse:
        get_logger().debug(f"ExitStepSync: {request.name}")
        await self._exit_barrier.wait()
        return sync_service.ExitStepSyncResponse()

    async def step(self, close: bool = False) -> bool:
        """
        Directly execute sync step without gRPC call

        - **Args**:
            - `close` (bool): Whether to close this participant

        - **Returns**:
            - bool: Whether this participant needs to be closed
        """
        # Enter sync state
        get_logger().debug("Virtual SyncerClient: Enter sync state")
        await self._enter_barrier.wait()

        # Exit sync state
        get_logger().debug("Virtual SyncerClient: Exit sync state")
        await self._exit_barrier.wait()

        return close
